from swibots import BotApp, RegisterCommand, BotContext, CommandEvent, InlineKeyboardButton, InlineMarkup, filters, CallbackQueryEvent
from tpblite import TPB
from logging import INFO, basicConfig, getLogger
from traceback import format_exc
import os                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        
BOT_TOKEN = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6MTA2MSwiaXNfYm90Ijp0cnVlLCJhY3RpdmUiOnRydWUsImlhdCI6MTY5MDgyNzYwMSwiZXhwIjoyMzIxOTc5NjAxfQ.A0Qv6mHgqbBoWdkWiKkc9e0boKwBHwY4QdTR7yOAxUA"                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        
basicConfig(                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            
    format="%(asctime)s  %(name)s [%(levelname)s] : %(message)s",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                     
    level=INFO,                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                         
    datefmt="%m/%d/%Y, %H:%M:%S",                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       
)                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       
LOGS = getLogger(__name__)                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        
try:                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    
    LOGS.info("Trying Connect With Switch")                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             
    LOGS.info("Trying Connect With Switch")
    bot = BotApp(BOT_TOKEN, "Torrent Searcher")
    LOGS.info("Successfully Connected with Switch")
    LOGS.info("Initialisation Torrent Searcher Client")
    torrent_searcher = TPB()
    LOGS.info("Initialised Torrent Searcher Client")
except Exception as e:
    LOGS.critical(str(e))
    exit()

# Registering Switch Commands

bot.register_command([
    RegisterCommand("torrents", "To Search On Torrents", True),
    RegisterCommand("besttorrent", "ToGet Best Torrent", True),
])

DATA = {}
FORMAT = """
*Torrent Title:* *{title}*

*Uploader:* `{uploader}`
*File Size:* `{filesize}`
*Number of Seeders:* `{seeders}`
*Number of Leechers:* `{leechers}`
"""

@bot.on_command("torrents")
async def src_tor(ctx: BotContext[CommandEvent]):
    event = ctx.event.message
    if not ctx.event.params.strip():
        return await event.reply_text("`Query Not Found!!`")
    xn = await event.reply_text("`Processing...`")
    torrents = torrent_searcher.search(ctx.event.params.strip())
    if len(torrents) == 0 or not torrents:
        return await xn.edit_text("`No Result Found!!!`")
    results = []
    for torrent in torrents:
        text = FORMAT.format(
            title=torrent.title,
            uploader=torrent.uploader,
            filesize=torrent.filesize,
            seeders=torrent.seeds,
            leechers=torrent.leeches
        )
        results.append([text, torrent.magnetlink])
    if len(results) > 1:
        DATA[int(event.id)] = results
        result = results[0]
        btn = [
            [InlineKeyboardButton("Manget Link 🔗", callback_data=f"link_0_{event.id}_{event.user_id}")],
            [
                # InlineKeyboardButton("<< Back", callback_data=f"back_0_{event.id}_{event.user_id}"),
                InlineKeyboardButton("Next >>", callback_data=f"next_0_{event.id}_{event.user_id}")
            ]
        ]
        return await xn.edit_text(result[0], inline_markup=InlineMarkup(inline_keyboard=btn))
    result = results[0]
    # mn = xn._prepare_response()
    inline_markup=InlineMarkup([[InlineKeyboardButton("Manget Link 🔗", callback_data=f"link_{id}_{event.id}_{event.user_id}")]])
    await event.edit_text(result[0], inline_markup=inline_markup)

@bot.on_command("besttorrent")
async def best_src_tor(ctx: BotContext[CommandEvent]):
    event = ctx.event.message
    if not ctx.event.params.strip():
        return await event.reply_text("`Query Not Found!!`")
    xn = await event.reply_text("`Processing...`")
    torrents = torrent_searcher.search(ctx.event.params.strip())
    torrent  = torrents.getBestTorrent()
    text = FORMAT.format(
        title=torrent.title,
        uploader=torrent.uploader,
        filesize=torrent.filesize,
        seeders=torrent.seeds,
        leechers=torrent.leeches
    )
    # await xn.edit_text(text, inline_markup=InlineMarkup([[InlineKeyboardButton("Manget Link 🔗", callback_data=f"link_{id}_{msg_id}_{user_id}")]]))


@bot.on_callback_query(filters.regexp(r"back_"))
async def _bck(ctx: BotContext[CallbackQueryEvent]):
    query = ctx.event.callback_data
    user = ctx.event.message.user.id
    spl = query.split("_")
    index, msg_id, user_id = int(spl[1]), int(spl[2]), int(spl[3])
    if int(user) != user_id:
        return
    if index > 0:
        results = DATA[msg_id]
        id  = index - 1
        if id == 0:
            result = results[id]
            btn = [
                [InlineKeyboardButton("Manget Link 🔗", callback_data=f"link_{id}_{msg_id}_{user_id}")],
                [
                    InlineKeyboardButton("Next >>", callback_data=f"next_{id}_{msg_id}_{user_id}")
                ]
            ]
        else:
            result = results[id]
            btn = [
                [InlineKeyboardButton("Manget Link 🔗", url= result[1])],
                [
                    InlineKeyboardButton("<< Back", callback_data=f"back_{id}_{msg_id}_{user_id}"),
                    InlineKeyboardButton("Next >>", callback_data=f"next_{id}_{msg_id}_{user_id}")
                ]
            ]
        # msg = ctx.event.message
        # mn = msg._prepare_response()
        # mn.message= result[0]
        # mn.inline_markup = InlineMarkup(btn)
#        return await bot.edit_message(mn)
        return await ctx.event.message.edit_text(result[0], InlineMarkup(inline_keyboard=btn))


@bot.on_callback_query(filters.regexp(r"next_"))
async def _nxt(ctx: BotContext[CallbackQueryEvent]):
    query = ctx.event.callback_data
    user = ctx.event.message.user.id
    spl = query.split("_")
    index, msg_id, user_id = int(spl[1]), int(spl[2]), int(spl[3])
    if int(user) != user_id:
        return print(user, user_id)
    results = DATA[msg_id]
    id  = index + 1
    result = results[id]
    if results[-1] == result:
        btn = [
            [InlineKeyboardButton("Manget Link 🔗", callback_data=f"link_{id}_{msg_id}_{user_id}")],
            [
                InlineKeyboardButton("<< Back", callback_data=f"back_{id}_{msg_id}_{user_id}"),
            ]
        ]
    else:
        btn = [
            [InlineKeyboardButton("Manget Link 🔗", callback_data=f"link_{id}_{msg_id}_{user_id}")],
            [
                InlineKeyboardButton("<< Back", callback_data=f"back_{id}_{msg_id}_{user_id}"),
                InlineKeyboardButton("Next >>", callback_data=f"next_{id}_{msg_id}_{user_id}")
            ]
        ]
    # msg = ctx.event.message
    # mn = msg._prepare_response()
    # mn.message= result[0]
    # mn.inline_markup = InlineMarkup(btn)
    # return await bot.edit_message(mn)
    return await ctx.event.message.edit_text(result[0], InlineMarkup(inline_keyboard=btn))

@bot.on_callback_query(filters.regexp(r"link_"))
async def _lnk(ctx: BotContext[CallbackQueryEvent]):
    query = ctx.event.callback_data
    user = ctx.event.message.user.id
    spl = query.split("_")
    index, msg_id, user_id = int(spl[1]), int(spl[2]), int(spl[3])
    if int(user) != user_id:
        return print(user, user_id)
    results = DATA[msg_id]
    result = results[index]
    btn = [[
        InlineKeyboardButton("<< Back", callback_data=f"back_{index+1}_{msg_id}_{user_id}")
    ]]
    # msg = ctx.event.message
    # mn = msg._prepare_response()
    # mn.message= result[1]
    # mn.inline_markup = InlineMarkup(btn)
    # return await bot.edit_message(mn)
    return await ctx.event.message.edit_text(result[1], InlineMarkup(inline_keyboard=btn))


bot.run()


import swibots
from typing import Optional, Any
from swibots.base import SwitchObject
from swibots.api.common.models import User
from swibots.utils.types import JSONDict


class Referral(SwitchObject):
    def __init__(
        self,
        app: "swibots.App" = None,
        id: Optional[str] = None,
        reference_id: Optional[str] = None,
        reference_type: Optional[str] = None,
        referral_code_id: Optional[str] = None,
        refer_by: Optional[str] = None,
        refer_to: Optional[str] = None,
        status: Optional[str] = None,
        user: Optional[User] = None,
    ):
        super().__init__(app)
        self.id = id
        self.reference_id = reference_id
        self.refer_to = refer_to
        self.refer_by = refer_by
        self.referral_code_id = referral_code_id
        self.status = status
        self.user = user
        self.reference_type = reference_type

    def to_json(self) -> JSONDict:
        return {
            "id": self.id,
            "reference_id": self.reference_id,
            "refer_to": self.refer_to,
            "refer_by": self.refer_by,
            "referral_code_id": self.referral_code_id,
            "user": self.user.to_json() if self.user else None,
            "reference_type": self.reference_type,
            "status": self.status,
        }

    @classmethod
    def from_json(self, data: JSONDict = None) -> Any:
        if data:
            self.id = data.get("id")
            self.reference_id = data.get("reference_id")
            self.refer_to = data.get("refer_to")
            self.refer_by = data.get("refer_by")
            self.user = User.build_from_json(data.get("user"), self.app)
            self.referral_code_id = data.get("referral_code_id")
            self.reference_type = data.get("reference_type")
            self.status = data.get("status")
        return self



import asyncio
import aiofiles
import base64
import hashlib
import httpx
import json
import logging
import mimetypes
import os
import random
import re
import tempfile
import uuid
from urllib.parse import urlparse
from datetime import datetime
from httpx import AsyncClient
from io import BytesIO
from typing import TYPE_CHECKING, Optional, Union
from typing import TYPE_CHECKING, Optional
from swibots.api.common.models import Media
from swibots.config import APP_CONFIG
from swibots.errors import UnknownBackBlazeError, FileTooLarge
from swibots.types import MAX_FILE_SIZE

from swibots.utils.types import (
    DownloadProgressCallback,
    IOClient,
    UploadProgress,
    UploadProgressCallback,
)

from swibots.errors import UnknownBackBlazeError, FileTooLarge
from swibots.utils.types import (
    UploadProgressCallback,
    DownloadProgressCallback,
    IOClient,
    UploadProgress,
)
from swibots.types import MAX_FILE_SIZE
from swibots.config import APP_CONFIG
from swibots.api.common.models import Media

if TYPE_CHECKING:
    from swibots.api.chat import ChatClient

log = logger = logging.getLogger(__name__)

BASE_PATH = "/v1/media"

account_id = APP_CONFIG["BACKBLAZE"].get("ACCOUNT_ID")
application_key = APP_CONFIG["BACKBLAZE"].get("APPLICATION_KEY")

bucket_id = APP_CONFIG["BACKBLAZE"].get("BUCKET_ID")

headers = {}
headers["Authorization"] = (
    "Basic " + base64.b64encode(f"{account_id}:{application_key}".encode()).decode()
)
headers["accept"] = "application/json"

MAX_THUMB_SIZE = 1024 * 40
logging.getLogger("httpx").setLevel(logging.ERROR)


class ReadableFile:
    def __init__(self, file, name, start_bytes: int = 0, max_read: int = None):
        self.file = file
        self.name = name
        self.readed = 0
        self.file.seek(start_bytes)
        self.max_read_size = max_read

    def read(self, n):
        if self.max_read_size and self.readed > self.max_read_size:
            return b""

        chunk = self.file.read(n)
        self.readed += len(chunk)
        if self.max_read_size and self.readed > self.max_read_size:
            chunk = chunk[: self.max_read_size - self.readed]
        return chunk

    def close(self):
        if hasattr(self.file, "close"):
            self.file.close()


class MediaService:
    def __init__(self) -> None:
        pass

    async def get_thumb_url(
        self, path: Union[str, BytesIO], for_document: bool = False, *args, **kwargs
    ) -> str:
        if not path:
            return

        if path and isinstance(path, str):
            # check for valid urls
            parse = urlparse(path)
            if parse.scheme and parse.netloc:
                return path

        __remove = False
        mime_type = (
            mimetypes.guess_type(path.name if isinstance(path, BytesIO) else path)[0]
            or "application/octet-stream"
        )
        hw = 30 if for_document else 100
        if "video/" in mime_type:
            path = await self.generate_from_ffmpeg(path, hw)
            if not path:
                return
            __remove = True
        elif not "image/" in mime_type:
            return

        content = None
        if isinstance(path, str):
            size = os.path.getsize(path)
            name = path
        else:
            size = path.getbuffer().nbytes
            name = path.name

        if size > MAX_THUMB_SIZE:
            try:
                from PIL import Image

                log.info(f"creating thumb for {path}")

                img = Image.open(content or path)
                img.thumbnail((hw, hw))
                path = tempfile.gettempdir() + f"/{uuid.uuid1()}.png"
                __remove = True
                img.save(path)
            except ImportError:
                log.warning(
                    "[Pillow is not installed] and [thumb size is greater than 20kb], ignoring thumb"
                )
                return

        _, ext = os.path.splitext(name)
        file_name = f"{uuid.uuid1()}{ext}"

        output = await self.file_to_response(
            path, mime_type, file_name, content=content, remove=__remove
        )
        if isinstance(output, dict):
            return output["downloadUrl"]

        path, file = output
        return f"{self.url}/file/{self.bucket_name}/{file['fileName']}"

    async def generate_from_ffmpeg(self, path: str, hw: int):
        log.debug("checking for ffmpeg")
        ffmpeg_path = os.getenv("FFMPEG_PATH") or "ffmpeg"
        try:
            proc = await asyncio.create_subprocess_exec(
                ffmpeg_path, stderr=asyncio.subprocess.PIPE
            )
            await proc.wait()
        except Exception as er:
            logger.error(er)
            return
        if b"ffmpeg version" not in (er := await proc.stderr.read()):
            log.info(f"ffmpeg is not installed, {er}")
            return
        name = os.path.join(tempfile.gettempdir(), f"{uuid.uuid1()}.png")
        log.info(f"generating thumb for '{path}'")
        cmd = [
            ffmpeg_path,
            "-i",
            path,
            "-ss",
            "00:00:01.000",
            "-vf",
            f"scale={hw}:-1:force_original_aspect_ratio=increase",
            "-vframes",
            "1",
            name,
        ]
        log.debug(f"Running command, {cmd}")
        proc = await asyncio.create_subprocess_shell(
            " ".join(cmd), stderr=asyncio.subprocess.PIPE
        )
        await proc.wait()
        if os.path.exists(name):
            return name
        log.info(await proc.stderr.read())


class StorageMediaService(MediaService):
    def __init__(self, client: "ChatClient") -> None:
        self.client = client
        self.api_url = "https://storage.switch.click"

        self._min_file_size = 200 * 1024 * 1024

    async def get_upload_url(self):
        async with AsyncClient(timeout=None) as client:
            response = await client.get(f"{self.api_url}/get_upload_url")
            return response.json()["url"]

    async def check_if_active(self):
        try:
            async with AsyncClient(timeout=None) as client:
                resp = await client.get(self.api_url)
                assert resp.status_code == 200
                return True
        except Exception as e:
            log.error(f"Failed to check if api is active: {e}")
            return False

    async def start_large_file(self, file_name: str, file_size: int):
        logger.info(f"Starting large file upload for {file_name} with size {file_size}")

        upload_url = await self.get_upload_url()
        async with AsyncClient(timeout=None) as client:
            response = await client.post(
                f"{upload_url}/start_large_file",
                json={"file_name": os.path.basename(file_name), "file_size": file_size},
            )
        response_data = response.json()
        response_data["upload_url"] = upload_url
        return response_data

    async def upload_part(
        self,
        upload_url: str,
        path: str,
        fileId: str,
        start_bytes: int,
        part_size: int,
        part_number: int,
        progress: UploadProgress,
        retries: int = 10,
    ):
        if isinstance(path, str):
            path = open(path, "rb")

        file = ReadableFile(path, path, start_bytes=start_bytes, max_read=part_size)

        async def __upload_part():
            for i in range(retries):
                try:
                    async with AsyncClient(timeout=None) as client:
                        response = await client.post(
                            "{}/upload_part?fileId={}&partNumber={}".format(
                                upload_url, fileId, part_number
                            ),
                            files={
                                "file": file,
                            },
                        )
                        return response.json()
                except Exception as e:
                    log.error(f"Failed to upload part {i}: {e}")
                    await asyncio.sleep(1)

            raise Exception(f"Failed to upload part {part_number} after 10 attempts")

        try:
            task = asyncio.create_task(__upload_part())

            readed = 0
            if progress.callback:
                while not task.done():
                    await asyncio.sleep(0.1)
                    if readed != file.readed:
                        change_read = file.readed - readed
                        if change_read + progress.readed <= progress.total:
                            await progress.bytes_readed(change_read)
                        readed = file.readed
            else:
                await task
        finally:
            file.close()

        return task.result()

    async def upload_media(
        self,
        path: str | BytesIO,
        caption: Optional[str] = None,
        file_name: Optional[str] = None,
        description: Optional[str] = None,
        thumb: Optional[str] = None,
        mime_type: Optional[str] = None,
        media_type: Optional[int] = None,
        callback: UploadProgressCallback = None,
        callback_args: Optional[tuple] = None,
        auto_thumb: Optional[bool] = True,
        part_size: int = int(os.getenv("UPLOAD_PART_SIZE", 200 * 1024 * 1024)),
        task_count: int = int(os.getenv("UPLOAD_TASKS", 0)),
        for_document: bool = False,
        premium: bool = False,
        retries: int = 10,
    ) -> Media:
        _is_bytesio = isinstance(path, BytesIO)
        if not part_size or part_size < self._min_file_size:
            part_size = self._min_file_size

        if not _is_bytesio:
            size = os.path.getsize(path)
        else:
            size = path.getbuffer().nbytes

        if size > MAX_FILE_SIZE:
            raise FileTooLarge(f"{path}: file size is too big to upload!")

        if not mime_type:
            mime_type = (
                mimetypes.guess_type(path.name if isinstance(path, BytesIO) else path)[
                    0
                ]
                or "application/octet-stream"
            )

        if size <= self._min_file_size:
            response = await self.file_to_response(
                path=path,
                callback=callback,
                callback_args=callback_args,
                file_size=size,
                mime_type=mime_type,
            )
            downloadUrl = response["downloadUrl"]
        else:
            if not file_name:
                file_name = os.path.basename(path)

            response = await self.start_large_file(file_name, size)
            upload_url = response["upload_url"]

            queue = asyncio.Queue()
            progress = UploadProgress(
                path=path,
                callback=callback,
                callback_args=callback_args,
                client=IOClient(),
                size=size,
            )
            upl_size = 0
            part_number = 0
            while upl_size < size:
                await queue.put(
                    self.upload_part(
                        path=path,
                        upload_url=upload_url,
                        start_bytes=upl_size,
                        part_size=part_size,
                        fileId=response["fileId"],
                        part_number=part_number,
                        progress=progress,
                        retries=retries,
                    )
                )
                upl_size += part_size
                part_number += 1

            logger.info(f"total parts: {part_number}")

            async def runFromQueue():
                while not queue.empty():
                    try:
                        task = await queue.get()
                        await task
                    except Exception as er:
                        log.exception(er)

            qTask = [asyncio.create_task(runFromQueue()) for _ in range(task_count)]

            try:
                await asyncio.wait(qTask)
            except Exception as er:
                log.exception(er)

            for q in qTask:
                if q.done() and (exc := q.exception()):
                    log.exception(exc)
                if not q.done():
                    q.cancel()

            response = await self.complete_large_file(
                upload_url, response["fileId"], part_number
            )
            downloadUrl = response["downloadUrl"]

        try:
            thumbUrl = await self.get_thumb_url(
                thumb or (path if auto_thumb else None), for_document
            )
        except Exception as er:
            log.exception(er)
            thumbUrl = None

        media = {
            "caption": caption,
            "description": description,
            "mimeType": mime_type,
            "fileSize": size,
            "fileName": file_name or os.path.basename(path),
            "downloadUrl": downloadUrl,
            "thumbnailUrl": thumbUrl,
            "mediaType": media_type,
            "sourceUri": response["fileId"],
            "checksum": " ",
            "ownerId": self.client.app.user.id,
            "premium": premium,
        }

        return self.client.build_object(Media, media)

    async def complete_large_file(
        self, upload_url: str, file_id: str, total_parts: int
    ):
        async with AsyncClient(timeout=None) as client:
            response = await client.post(
                f"{upload_url}/finish_large_file",
                json={"totalParts": total_parts, "fileId": file_id, "blocking": True},
            )
            return response.json()

    async def file_to_response(
        self,
        path: str | BytesIO,
        mime_type=None,
        file_name=None,
        content: bytes = None,
        callback=None,
        callback_args=None,
        file_size=None,
        remove: bool = None,
    ):
        upload_url = await self.get_upload_url()
        Isbytes = isinstance(path, BytesIO)

        if not Isbytes:
            file = open(path, "rb")
        file = ReadableFile(file, path.name if Isbytes else path)

        if not mime_type:
            mime_type = (
                mimetypes.guess_type(path.name if Isbytes else path)[0]
                or "application/octet-stream"
            )

        async def upload_file():
            async with AsyncClient(timeout=None) as client:
                response = await client.post(
                    f"{upload_url}/upload?blocking=true", files={"file": file}
                )
                jsonData = response.json()

            file.close()
            return jsonData

        task = asyncio.create_task(upload_file())

        if callback:
            readed = 0
            while not task.done():
                await asyncio.sleep(0.1)
                if readed != file.readed:
                    change_read = file.readed - readed
                    readed = file.readed
                    progress = UploadProgress(
                        path=path,
                        callback=callback,
                        callback_args=callback_args,
                        client=IOClient(),
                        size=file_size,
                    )
                    if change_read + progress.readed <= progress.total:
                        await progress.bytes_readed(change_read)
        else:
            await task

        result = task.result()
        if remove and not Isbytes:
            os.remove(path)

        return result


class MediaController(MediaService):
    """Media controller
    This controller is used to communicate with the media endpoints.
    """

    MIN_WAIT = 3
    MAX_WAIT = 7

    def __init__(self, client: "ChatClient", bucket_name: str = "lambooo"):
        self.client = client
        self.__token = None
        self.url = None
        self._min_part_size = 5000000
        self.bucket_name = bucket_name
        self.storage_service = StorageMediaService(client)

    async def getAccountInfo(self):
        response = await self.request(
            "https://api.backblazeb2.com/b2api/v2/b2_authorize_account",
            headers=headers,
            method="GET",
        )
        data = response.json()
        self.url = data["apiUrl"]

        if token := data.get("authorizationToken"):
            self.__token = token
        log.debug(data)
        if min_size := data.get("absoluteMinimumPartSize"):
            self._min_part_size = min_size

        return self.__token

    async def file_to_response(
        self,
        path: str | BytesIO,
        mime_type=None,
        file_name=None,
        content: bytes = None,
        callback=None,
        callback_args=None,
        file_size=None,
        remove: bool = None,
    ):
        token = self.__token or await self.getAccountInfo()
        Isbytes = isinstance(path, BytesIO)

        if not mime_type:
            mime_type = (
                mimetypes.guess_type(path.name if Isbytes else path)[0]
                or "application/octet-stream"
            )

        head = {
            "Content-Type": "application/json",
            "Authorization": token,
        }
        rsp = await self.request(
            f"{self.url}/b2api/v2/b2_get_upload_url?bucketId={bucket_id}",
            method="GET",
            headers=head,
        )

        response_data = rsp.json()
        if not response_data.get("authorizationToken"):
            raise UnknownBackBlazeError(response_data)
        token = response_data["authorizationToken"]

        respp = rsp.json()
        if not respp.get("uploadUrl"):
            raise UnknownBackBlazeError(respp)

        with path if Isbytes else open(path, "rb") as f:
            content = f.read()
            file_sha1 = hashlib.sha1(content).hexdigest()

            headers = {
                "Authorization": token,
                "X-Bz-File-Name": file_name,
                "Content-Type": mime_type,
                "X-Bz-Content-Sha1": file_sha1,
                "Content-Length": str(len(content)),
            }

            rsp = await self.request(
                respp["uploadUrl"],
                headers=headers,
                data=content,
            )
            if Isbytes:
                path = BytesIO(content)
                path.name = file_name
        if remove:
            os.remove(path)

        file_response = rsp.json()
        if "contentLength" in file_response:
            if callback:
                progress = UploadProgress(
                    path=path,
                    callback=callback,
                    callback_args=callback_args,
                    client=IOClient(),
                    size=file_size,
                )
                await progress.bytes_readed(file_response["contentLength"])
        else:
            log.error(file_response)
        return path, file_response

    async def upload_media(
        self,
        path: str | BytesIO,
        caption: Optional[str] = None,
        file_name: Optional[str] = None,
        description: Optional[str] = None,
        thumb: Optional[str] = None,
        mime_type: Optional[str] = None,
        media_type: Optional[int] = None,
        callback: UploadProgressCallback = None,
        callback_args: Optional[tuple] = None,
        auto_thumb: Optional[bool] = True,
        part_size: int = int(os.getenv("UPLOAD_PART_SIZE", 0)),
        task_count: int = int(os.getenv("UPLOAD_TASKS", 0)),
        min_file_size: int = None,
        for_document: bool = False,
        premium: bool = False,
        retries: int = 10,
        private_community: bool = False,
        storage_method_retries: int = 3,
    ) -> Media:
        """Upload media to switch

        Args:
            path (str | BytesIO): path to upload
            caption (Optional[str], optional): caption of media Defaults to None.
            description (Optional[str], optional): file name of media. Defaults to None.
            thumb (Optional[str], optional): path to use as thumb for media. Defaults to None.
            mime_type (Optional[str], optional): MIME type of media. Defaults to None.
            media_type (Optional[int], optional): Media type. Defaults to None.
            callback (UploadProgressCallback, optional): Callback progress. Defaults to None.
            callback_args (Optional[tuple], optional): Additional args to use in callbacks. Defaults to None.
            part_size (int, optional): part size while uploading. Defaults to None.
            task_count (int, optional): number of tasks while uploading, Defaults to None.

        Returns:
            Media:
        """
        if not task_count:
            task_count = 1

        if not file_name:
            if isinstance(path, BytesIO):
                file_name = path.name
            else:
                file_name = path

        if private_community or self.client.app._upload_mode == "STORAGE":
            if await self.storage_service.check_if_active():
                for _ in range(storage_method_retries):
                    if _:
                        logger.info(f"Retrying storage method: {_}")
                    try:
                        return await self.storage_service.upload_media(
                            path=path,
                            caption=caption,
                            file_name=file_name,
                            description=description,
                            thumb=thumb,
                            mime_type=mime_type,
                            media_type=media_type,
                            callback=callback,
                            callback_args=callback_args,
                            auto_thumb=auto_thumb,
                            part_size=part_size,
                            task_count=task_count,
                            premium=premium,
                            retries=storage_method_retries,
                            for_document=for_document,
                        )
                        break
                    except Exception as er:
                        logger.error(f"Error in storage method: {er}")
                        logger.exception(er)

        if not part_size:
            part_size = self._min_part_size

        if not min_file_size:
            min_file_size = self._min_part_size

        if part_size < self._min_part_size:
            log.warning(
                f"part_size cant be smaller than minimum [{self._min_part_size}]"
            )
            part_size = self._min_part_size

        if not mime_type:
            mime_type = mimetypes.guess_type(file_name)[0] or "application/octet-stream"
        log.debug(f"Sending request to backblaze: {path}")
        _is_bytesio = isinstance(path, BytesIO)

        if not _is_bytesio:
            size = os.path.getsize(path)
        else:
            size = path.getbuffer().nbytes
        if size > MAX_FILE_SIZE:
            raise FileTooLarge(f"{path}: file size is too big to upload!")
        _, ext = os.path.splitext(path.name if _is_bytesio else path)
        file_name = f"{uuid.uuid1()}{ext}"

        if (
            size > min_file_size
            and size > self._min_part_size
            and (size // part_size) > 1
        ):
            file_response = await self.upload_large_file(
                path,
                callback=callback,
                content_type=mime_type,
                file_name=file_name,
                callback_args=callback_args,
                part_size=part_size,
                task_count=task_count,
                file_size=size,
                file_info={
                    "timestamp": str(datetime.now().timestamp()),
                    "uploaded_by": str(self.client.user.id),
                },
                retries=retries,
            )
        else:
            path, file_response = await self.file_to_response(
                path,
                mime_type,
                file_name,
                callback=callback,
                file_size=size,
                callback_args=callback_args,
            )
        if not file_response.get("fileName"):
            raise UnknownBackBlazeError(file_response)
        url = f"{self.url}/file/{self.bucket_name}/{file_response['fileName']}"
        try:
            thumbUrl = await self.get_thumb_url(
                thumb or (path if auto_thumb else None), for_document
            )
        except Exception as er:
            log.exception(er)
            thumbUrl = None
        media = {
            "caption": caption,
            "description": description,
            "mimeType": mime_type,
            "fileSize": file_response.get("size", size),
            "fileName": file_response["fileName"],
            "downloadUrl": url,
            "thumbnailUrl": thumbUrl,
            "mediaType": media_type,
            "sourceUri": file_response["fileId"],
            "checksum": file_response["contentSha1"],
            "ownerId": self.client.app.user.id,
            "premium": premium,
        }
        return self.client.build_object(Media, media)

    async def request(self, url: str, method: str = "POST", **kwargs):
        async with AsyncClient(
            verify=False, timeout=None, headers=kwargs.get("headers")
        ) as client:
            resp = await (client.post if method == "POST" else client.get)(
                url, **kwargs
            )
            message = resp.json()
            if message.get("code") == "expired_auth_token":
                log.error("Expired auth token, retrying")
                token = await self.getAccountInfo()
                headers = kwargs.get("headers") or {}
                headers["Authorization"] = token
                kwargs["headers"] = headers
                return await self.request(url, method, **kwargs)
            elif message.get("code") == "service_unavailable":
                log.error(message)
                randomTime = random.randint(self.MIN_WAIT, self.MAX_WAIT)
                log.info(f"Waiting for {randomTime} before retry!")
                await asyncio.sleep(randomTime)
                return await self.request(url, method, **kwargs)
            return resp

    async def __upload_file(
        self,
        token,
        part_number,
        path,
        upl_size,
        part_size,
        fileId: str,
        progress: UploadProgress,
        retries: int = 10,
        wait_factor: int = 6,
    ):
        async with aiofiles.open(path, "rb") as f:
            for _ in range(retries):
                try:
                    respp = await self.request(
                        f"{self.url}/b2api/v2/b2_get_upload_part_url",
                        json={"fileId": fileId},
                        headers={"Authorization": token},
                        timeout=30,
                    )
                    break
                except Exception as er:
                    log.debug(er)
                    log.error(er)
                    if _ == (retries - 1):
                        raise er

            resp_data = respp.json()
            if respp.status_code != 200:
                logger.error("on Part url")
                logger.error(resp_data)
                raise UnknownBackBlazeError(resp_data)

            token = resp_data["authorizationToken"]
            upload_part_url = resp_data["uploadUrl"]

            await f.seek(upl_size)
            chunk = await f.read(part_size)

            for _ in range(retries + 1):
                if _:
                    log.info(f"Retrying upload [{path}][{upl_size}:{part_size}]")
                try:
                    respp = await self.request(
                        upload_part_url,
                        data=chunk,
                        headers={
                            "Authorization": token,
                            "X-Bz-Part-Number": str(part_number),
                            "X-Bz-Content-Sha1": hashlib.sha1(chunk).hexdigest(),
                        },
                    )
                    resp_data = respp.json()
                    if resp_data.get("code") == "service_unavailable":
                        log.error(resp_data)
                        log.info(f"Waiting for {_ * wait_factor}")
                        await asyncio.sleep(_ * wait_factor)
                        continue
                    if respp.status_code != 200:
                        logger.error("onUpload")
                        logger.error(resp_data)
                    hash = resp_data["contentSha1"]
                    await progress.bytes_readed(resp_data["contentLength"])
                    return hash, resp_data["partNumber"]
                except (httpx.WriteError, httpx.ReadError) as er:
                    log.exception(er)
                    log.error(er.request)
                except Exception as er:
                    log.exception(er)

    async def upload_large_file(
        self,
        path: str | BytesIO,
        callback,
        content_type,
        file_name,
        file_info=None,
        callback_args=None,
        part_size=None,
        task_count=None,
        file_size=None,
        retries: int = None,
    ):
        log.info("getting account info")
        token = await self.getAccountInfo()

        progress = UploadProgress(
            path=path,
            callback=callback,
            callback_args=callback_args,
            client=IOClient(),
            size=file_size,
        )
        if isinstance(path, BytesIO) and not file_name:
            file_name = path.name
        upl_size = 0
        head = {
            "Content-Type": content_type,
            "X-Bz-File-Name": file_name,
            "Authorization": token,
        }
        data = {
            "fileName": file_name,
            "contentType": content_type,
            "bucketId": bucket_id,
            "fileInfo": file_info,
        }
        partHash = {}
        logger.info("start large file")
        respp = await self.request(
            f"{self.url}/b2api/v2/b2_start_large_file",
            headers=head,
            data=json.dumps(data),
        )
        if respp.status_code != 200:
            logger.error("on large file")
            logger.error(respp.json())
            raise UnknownBackBlazeError(respp.json())

        logger.debug(respp.json())

        fileId = respp.json()

        if not fileId.get("fileId"):
            raise UnknownBackBlazeError(fileId)

        fileId = fileId["fileId"]
        part_number = 1
        queue = asyncio.Queue()

        async def runFromQueue():
            while queue.qsize():
                task = await queue.get()
                try:
                    hash, part_number = await task
                except Exception as er:
                    log.exception(er)
                    continue
                partHash[hash] = part_number

        try:
            while upl_size < file_size:
                await queue.put(
                    self.__upload_file(
                        token,
                        part_number,
                        path,
                        upl_size,
                        part_size,
                        fileId,
                        progress,
                        retries=retries,
                    )
                )
                part_number += 1
                upl_size += part_size
        except Exception as er:
            log.exception(er)
        qTask = [asyncio.create_task(runFromQueue()) for _ in range(task_count)]

        try:
            await asyncio.wait(qTask)
        except Exception as er:
            log.exception(er)

        for q in qTask:
            if q.done() and (exc := q.exception()):
                log.exception(exc)
            if not q.done():
                q.cancel()

        if not partHash:
            raise Exception("parts are not found!")
        response = await self.__finish_large_file(
            partHash, fileId, path, part_size=part_size, progress=progress, token=token
        )
        return response

    async def __finish_large_file(
        self, partHash, fileId, path, part_size, progress, token, retry_count: int = 0
    ):
        if retry_count > 3:
            raise Exception("Max retries reached for finish file")
        hashes = list(map(lambda x: x[0], sorted(partHash.items(), key=lambda x: x[1])))
        try:
            response = await self.request(
                f"{self.url}/b2api/v2/b2_finish_large_file",
                json={
                    "fileId": fileId,
                    "partSha1Array": hashes,
                },
                headers={"Authorization": token},
            )
        except Exception as er:
            log.error("Error on finish large file")
            log.exception(er)
            log.info("canceling upload")
            resp = await self.request(
                f"{self.url}/b2api/v2/b2_cancel_large_file",
                data={"fileId": fileId},
                headers={"Authorization": token},
            )
            log.info(resp)
            log.info(resp.json())
        response = response.json()
        if response.get("code") == "bad_request":
            logger.info(response)
            mtch = re.search(
                "Part number (\d+) has not been uploaded", response.get("message")
            )
            if mtch:
                part_number = int(mtch.group(1))
                hash, part_number = await self.__upload_file(
                    token,
                    part_number,
                    path,
                    fileId=fileId,
                    upl_size=part_size * (part_number - 1),
                    part_size=part_size,
                    progress=progress,
                )
                partHash[hash] = part_number
                retry_count += 1
                return await self.__finish_large_file(
                    partHash,
                    fileId,
                    path,
                    part_size,
                    progress,
                    token=token,
                    retry_count=retry_count,
                )
        return response

    async def get_media(self, media_id: int) -> Media:
        response = await self.client.get(f"{BASE_PATH}/{media_id}")
        return self.client.build_object(Media, response.data)

    async def update_media(self, media_id: int, media: Optional[Media] = None) -> Media:
        data = media.to_update_request()

        data["id"] = media_id
        data = {x: y for x, y in data.items() if y}
        response = await self.client.put(
            BASE_PATH,
            data=data,
        )

        return self.client.build_object(Media, response.data)

    async def delete_media(self, media_id: int):
        await self.client.delete(
            "/v1/community/channel/media", data={"mediaId": media_id}
        )
        return True

    async def download_by_media_id(
        self,
        media_id: int,
        file_name: str = None,
        in_memory: Optional[bool] = False,
        block: Optional[bool] = True,
        progress: Optional[DownloadProgressCallback] = None,
        progress_args: Optional[tuple] = (),
    ):
        url = f"{BASE_PATH}/{media_id}/download"
        return await self.client.app.handle_download(
            url,
            file_name=file_name,
            in_memory=in_memory,
            block=block,
            progress=progress,
            progress_args=progress_args,
        )



from typing import Any, List, Optional
import swibots
from swibots.base import SwitchObject

from swibots.utils.types import JSONDict, SCT
from .inline_query_result import InlineQueryResult


class InlineQueryAnswer(SwitchObject):
    def __init__(
        self,
        app: "swibots.App" = None,
        query_id: Optional[str] = None,
        user_id: int = None,
        title: Optional[str] = None,
        results: List["InlineQueryResult"] = None,
        next_offset: Optional[str] = None,
        cache_time: int = None,
        is_personal: bool = True,
        pm_text: Optional[str] = None,
        pm_parameter: Optional[str] = None,
    ):
        super().__init__(app)
        self.query_id = query_id
        self.user_id = user_id
        self.title = title
        self.results = results
        self.next_offset = next_offset
        self.cache_time = cache_time
        self.is_personal = is_personal
        self.pm_text = pm_text
        self.pm_parameter = pm_parameter

    def to_json(self) -> JSONDict:
        return {
            "queryId": self.query_id,
            "userId": self.user_id,
            "title": self.title,
            "results": [self._result_json(result) for result in self.results or []],
            "nextOffset": self.next_offset,
            "cacheTime": self.cache_time,
            "isPersonal": self.is_personal,
            "pmText": self.pm_text,
            "pmParameter": self.pm_parameter,
        }

    def _result_json(self, result):
        if isinstance(result, dict):
            return result
        else:
            return result.to_json()

    def add_result(self, result: SCT[Any]) -> "InlineQueryAnswer":
        if not self.results:
            self.results = []

        if isinstance(result, List):
            self.results.extend(result)
        else:
            self.results.append(result)



from swibots.utils.types import JSONDict
from .types import SwitchObject, Component
from typing import List, Optional


class ListItem(Component):
    type = "list_tile"

    def __init__(
        self,
        title: str = None,
        subtitle: str = None,
        callback_data: str = None,
    ):
        self.title = title
        self.subtitle = subtitle
        self.callback_data = callback_data

    def to_json(self) -> JSONDict:
        data = {"type": self.type}
        if self.title:
            data["title"] = self.title
        if self.subtitle:
            data["subTitle"] = self.subtitle
        if self.callback_data:
            data["callbackData"] = self.callback_data
        return data



from .chat_event import ChatEvent
from .message_event import MessageEvent
from .command_event import CommandEvent
from .callback_query_event import CallbackQueryEvent
from .inline_query_event import InlineQueryEvent

__all__ = [
    "ChatEvent",
    "MessageEvent",
    "CommandEvent",
    "CallbackQueryEvent",
    "InlineQueryEvent",
]



from typing import Optional
from swibots.utils.types import JSONDict
from swibots.base.switch_object import SwitchObject


class EmbedInlineField(SwitchObject):
    def __init__(
        self,
        icon: Optional[str] = None,
        key: Optional[str] = None,
        title: Optional[str] = None,
    ):
        super().__init__()
        self.icon = icon
        self.key = key
        self.title = title

    def to_json(self) -> JSONDict:
        return {"icon": self.icon, "key": self.key, "title": self.title}



from .types import InlineQueryResultType
import swibots
from swibots.utils.types import JSONDict
from swibots.base import SwitchObject


class InputMessageContent(SwitchObject):
    def __init__(
        self,
        message_text: str = None,
    ):
        self.message_text = message_text

    def to_json(self) -> JSONDict:
        return {
            "messageText": self.message_text,
        }

    def from_json(self, data: JSONDict) -> "InputMessageContent":
        if data is not None:
            self.message_text = data.get("messageText")
        return self



from typing import Type, TypeVar
import swibots
from swibots.api.chat.models import Message


class FlagMessage:
    async def flag_message(self: "swibots.ApiClient", message: Message | int) -> bool:
        """Flag a message

        Parameters:
            message (``~switch.api.chat.models.Message`` | ``int``): The message to flag

        Returns:
            ``bool``: True if the message was flagged

        Raises:
            ``~switch.error.SwitchError``: If the message could not be flagged

        This method does the same as :meth:`~switch.api.chat.controllers.MessageController.flag_message`.
        """
        return await self.chat_service.messages.flag_message(message)



import swibots


class GetUnreadMessagesCount:
    async def get_unread_messages_count(self: "swibots.ApiClient") -> int:
        """Get the amount of unread messages

        Returns:
            ``int``: The amount of unread messages

        Raises:
            ``~switch.error.SwitchError``: If the amount of unread messages could not be retrieved

        This method does the same as :meth:`~switch.api.chat.controllers.MessageController.get_unread_messages_count`.
        """
        return await self.chat_service.messages.get_unread_messages_count()



from typing import Optional, List
from swibots.utils.types import JSONDict
from swibots.base.switch_object import SwitchObject
import swibots


class Quest(SwitchObject):
    def __init__(
        self,
        app: "swibots.App" = None,
        category: Optional[str] = None,
        category_id: Optional[str] = None,
        category_name: Optional[str] = None,
        claimed: Optional[bool] = None,
        community_id: Optional[str] = None,
        created_by: Optional[int] = None,
        description: Optional[str] = None,
        enabled: Optional[bool] = None,
        internal: Optional[bool] = None,
        name: Optional[str] = None,
        quest: Optional[str] = None,
        winner_ids: List[int] = None,
        xp: List[int] = None,
    ):
        super().__init__(app)
        self.category = category
        self.category_id = category_id
        self.category_name = category_name
        self.claimed = claimed
        self.community_id = community_id
        self.created_by = created_by
        self.description = description
        self.enabled = enabled
        self.internal = internal
        self.name = name
        self.quest = quest
        self.winner_ids = winner_ids
        self.xp = xp

    def to_json(self) -> JSONDict:
        return {
            "communityId": self.community_id,
            "enabled": self.enabled,
            "internal": self.internal,
            "quest": self.quest,
            "claimed": self.claimed,
            "winnerIds": self.winner_ids,
            "description": self.description,
            "category": self.category,
            "categoryName": self.category_name,
            "categoryId": self.category_id,
            "xp": self.xp,
            "createdBy": self.created_by,
        }

    def from_json(self, data: JSONDict = None) -> "Quest":
        if data is not None:
            self.community_id = data.get("communityId")
            self.enabled = data.get("enabled")
            self.internal = data.get("internal")
            self.name = data.get("name")
            self.quest = data.get("quest")
            self.claimed = data.get("claimed")
            self.winner_ids = data.get("winnerIds")
            self.description = data.get("description")
            self.category = data.get("category")
            self.category_id = data.get("categoryId")
            self.category_name = data.get("categoryName")
            self.xp = data.get("xp")
            self.created_by = int(data.get("createdBy") or 0)
        return self


class QuestCategory(SwitchObject):
    def __init__(
        self,
        app: "swibots.App" = None,
        category_id: Optional[str] = None,
        category_name: Optional[str] = None,
        community_id: Optional[str] = None,
        created_at: Optional[int] = None,
        updated_at: Optional[int] = None,
    ):
        super().__init__(app)
        self.community_id = community_id
        self.category_name = category_name
        self.created_at = created_at
        self.updated_at = updated_at
        self.category_id = category_id

    def from_json(self, data: JSONDict = None) -> "QuestCategory":
        if data is not None:
            self.category_id = data.get("categoryId")
            self.category_name = data.get("categoryName")
            self.community_id = data.get("communityId")
            self.created_at = data.get("createdAt")
            self.updated_at = data.get("updatedAt")
        return self



import json
import logging
from typing import TYPE_CHECKING, List
from swibots.api.airdrop.models import Referral, Tournament
from swibots.errors import SwitchError
from swibots.utils.types import JSONDict

if TYPE_CHECKING:
    from swibots.api.airdrop import AirdropClient

log = logging.getLogger(__name__)

BASE_PATH = "/v1/tournament"


class TournamentController:
    """Airdrop controller

    This controller is used to communicate with the airdrop endpoints.
    """

    def __init__(self, client: "AirdropClient"):
        self.client = client

    async def get_referrals(self, id: str | int, user_id: str | int) -> List[Referral]:
        """Get referrals"""
        response = await self.client.get(
            f"{BASE_PATH}/getUserReferrals/{id}?user_id={user_id}"
        )
        return self.client.build_list(Referral, response.data)

    async def get_tournaments(self, community_id: str) -> List[Tournament]:
        """Get tournaments by community_id"""
        response = await self.client.get(
            f"{BASE_PATH}/getTournamentsByCommunityId/{community_id}"
        )
        return self.client.build_list(Tournament, response.data)



from typing import Type, TypeVar, Optional, List
import swibots
from swibots.api.callback import AppPage


class AnswerCallback:
    async def answer_callback_query(
        self: "swibots.ApiClient",
        query_id: str,
        text: Optional[str] = None,
        url: Optional[str] = None,
        message_id: Optional[int] = None,
        show_alert: Optional[bool] = False,
        cache_time: Optional[int] = None,
        app_session_id: Optional[str] = None,
    ) -> bool:
        """
        Answer Callback query from callback button

        Arguments:
        query_id: str - Callback query id
        message_id: int:  message id linked to the message
        text: str - text
        url: str
        show_alert: bool - whether to display popup
        cache_time: int - cache time
        """
        return await self.bots_service.bots.answer_callback_query(
            query_id,
            text=text,
            url=url,
            show_alert=show_alert,
            cache_time=cache_time,
            message_id=message_id,
            app_session_id=app_session_id,
        )

    async def answer_ui_query(
        self: "swibots.ApiClient",
        query_id: str,
        message_id: int,
        callback: AppPage,
        app_session_id: str,
    ) -> bool:
        """Answer UI Query

        Args:
            query_id (str): callback query id
            message_id (int): message id linked to button
            callback (AppPage): Callback component.

        Returns:
            Bool: whether callback was sent or not
        """
        return await self.bots_service.bots.answer_ui_query(
            query_id, message_id, callback=callback, app_session_id=app_session_id
        )



from .community_event import CommunityEvent
from .channel_updated_event import ChannelUpdatedEvent
from .channel_created_event import ChannelCreatedEvent
from .channel_deleted_event import ChannelDeletedEvent
from .community_updated_event import CommunityUpdatedEvent
from .group_created_event import GroupCreatedEvent
from .group_updated_event import GroupUpdatedEvent
from .group_deleted_event import GroupDeletedEvent
from .member_joined_event import MemberJoinedEvent
from .member_left_event import MemberLeftEvent
from .user_banned_event import UserBannedEvent



from typing import Type, TypeVar
import swibots
from swibots.api.chat.models import InlineQueryAnswer, InlineQuery


class AnswerInlineQuery:
    async def answer_inline_query(
        self: "swibots.ApiClient", query: InlineQuery, answer: InlineQueryAnswer
    ) -> bool:
        """Answer inline query

        Args:
            query (:obj:`InlineQuery`): Inline query to answer to
            answer (str | InlineQueryAnswer): Answer to inline query, it can be either a string or an InlineQueryAnswer object, if it's a string, it will be used as the text of the answer
        """
        return await self.chat_service.messages.answer_inline_query(query, answer)



from typing import List, Union

from swibots.utils.types import JSONDict
from .types import Text, Icon, Component


class TableTile(Component):
    type = "table_tile"

    def __init__(
        self,
        title: Union[Text, str],
        icon: Union[Icon, str],
        subtitle: str = "",
        callback_data: str = None,
    ):
        if isinstance(title, str):
            title = Text(title)
        self.title = title

        if isinstance(icon, str):
            icon = Icon(icon)
        self.icon = icon
        self.subtitle = subtitle
        self.callback_data = callback_data

    def to_json(self):
        return {
            "title": self.title.to_json(),
            "icon": self.icon.to_json(),
            "subTitle": self.subtitle,
            "callbackData": self.callback_data,
            "type": self.type,
        }


class Table(Component):
    type = "table"

    def __init__(
        self,
        headings: List[TableTile],
        rows: List[List[TableTile]],
        title: str = "",
        scoreboard: bool = False,
        max_size: bool = None,
    ):
        self.title = title
        self.headings = headings
        self.rows = rows
        self.columns_count = len(headings)
        self.scoreboard = scoreboard
        self.max_size = max_size

    def to_json(self):
        return {
            "type": self.type,
            "headings": [heading.to_json() for heading in self.headings],
            "rows": [[row.to_json() for row in col] for col in self.rows],
            "columnsCount": self.columns_count,
            "version": 1 if self.scoreboard else 0,
            "mainAxisSize": "max" if self.max_size else "min",
        }



import logging
from typing import TYPE_CHECKING, List
from swibots.api.community.models import Group
from swibots.errors import SwitchError

if TYPE_CHECKING:
    from swibots.api.community import CommunityClient

log = logging.getLogger(__name__)

BASE_PATH = "/v1/community/group"


class GroupController:
    def __init__(self, client: "CommunityClient"):
        self.client = client

    async def get_group(self, group_id: str):
        """Get a channel by id"""
        response = await self.client.get(f"{BASE_PATH}?groupId={group_id}")
        return self.client.build_object(Group, response.data.get("result"))

    async def get_all_groups(self, community_id: str) -> List[Group]:
        response = await self.client.get(f"{BASE_PATH}/all?communityId={community_id}")
        return self.client.build_list(Group, response.data.get("result"))

    async def create_group(self, group: Group) -> str:
        response = await self.client.post(BASE_PATH, data=group.to_json_request())
        return response.data.get("result", {}).get("groupId")

    async def delete_group(self, group_id: str) -> bool:
        response = await self.client.delete(f"{BASE_PATH}/{group_id}")
        if response.status_code == 200:
            return True
        logging.debug(response.data)
        if err := response.data.get("errorMessage"):
            raise SwitchError(err)
        return False

    async def update_group(self, group: Group) -> str:
        response = await self.client.put(BASE_PATH, data=group.to_json_request())
        return response.data.get("result", {}).get("groupId")



import os


APP_CONFIG = {
    "CHAT_SERVICE": {
        "BASE_URL": os.getenv("CHAT_SERVICE_BASE_URL") or "https://chat-api.switch.pe",
        "WS_URL": os.getenv("CHAT_SERVICE_WS_URL")
        or "wss://chat-api.switch.pe/v1/websocket/message/ws",
    },
    "BOT_SERVICE": {
        "BASE_URL": os.getenv("BOT_SERVICE_BASE_URL") or "https://chat-api.switch.pe",
    },
    "AUTH_SERVICE": {
        "BASE_URL": os.getenv("AUTH_SERVICE_BASE_URL")
        or "https://gateway.switch.pe/user-service",
    },
    "AIRDROP_SERVICE": {
        "BASE_URL": os.getenv("AIRDROP_SERVICE_BASE_URL")
        or "https://gateway.switch.pe/airdrop-service"
    },
    "COMMUNITY_SERVICE": {
        "BASE_URL": os.getenv("COMMUNITY_SERVICE_BASE_URL")
        or "https://gateway.switch.pe/community-service",
        "WS_URL": os.getenv("COMMUNITY_SERVICE_WS_URL")
        or "wss://gateway.switch.pe/community-service/v1/websocket/community/ws",
    },
    "BACKBLAZE": {
        "BUCKET_ID": os.getenv("BACKBLAZE_BUCKET_ID") or "631c2d59c88c013e9721011f",
        "ACCOUNT_ID": os.getenv("BACKBLAZE_ACCOUNT_ID") or "3cd98c1e711f",
        "APPLICATION_KEY": os.getenv("BACKBLAZE_APPLICATION_KEY")
        or "005c20136383930ce1d7dc58757aa7fe646d64e712",
    },
}


def get_config():
    return APP_CONFIG


def reload_config():
    APP_CONFIG["CHAT_SERVICE"]["BASE_URL"] = (
        os.getenv("CHAT_SERVICE_BASE_URL") or "https://chat-api.switch.pe"
    )
    APP_CONFIG["CHAT_SERVICE"]["WS_URL"] = (
        os.getenv("CHAT_SERVICE_WS_URL")
        or "wss://chat-api.switch.pe/v1/websocket/message/ws"
    )
    APP_CONFIG["BOT_SERVICE"]["BASE_URL"] = (
        os.getenv("BOT_SERVICE_BASE_URL") or "https://chat-api.switch.pe"
    )
    APP_CONFIG["AUTH_SERVICE"]["BASE_URL"] = (
        os.getenv("AUTH_SERVICE_BASE_URL") or "https://gateway.switch.pe/user-service"
    )
    APP_CONFIG["COMMUNITY_SERVICE"]["BASE_URL"] = (
        os.getenv("COMMUNITY_SERVICE_BASE_URL")
        or "https://gateway.switch.pe/community-service"
    )
    APP_CONFIG["COMMUNITY_SERVICE"]["WS_URL"] = (
        os.getenv("COMMUNITY_SERVICE_WS_URL")
        or "wss://gateway.switch.pe/community-service/v1/websocket/community/ws"
    )
    APP_CONFIG["AIRDROP_SERVICE"]["BASE_URL"] = (
        os.getenv("AIRDROP_SERVICE_BASE_URL")
        or "https://gateway.switch.pe/airdrop-service"
    )


reload_config()


from typing import Optional
import swibots
from swibots.api.common.events.event import Event
from swibots.api.community.models.channel import Channel
from swibots.api.community.models.community import Community
from swibots.api.community.models.group import Group
from swibots.api.common.models.user import User
from swibots.api.chat.models.message import Message
from swibots.api.chat.models.inline.inline_query import InlineQuery
from swibots.types import EventType
from swibots.utils.types import JSONDict
from .chat_event import ChatEvent


class InlineQueryEvent(ChatEvent):
    """Message event"""

    def __init__(
        self,
        app: "swibots.App" = None,
        type: Optional[EventType] = None,
        community_id: Optional[str] = None,
        community: Optional[Community] = None,
        group_id: Optional[str] = None,
        group: Optional[Group] = None,
        channel_id: Optional[str] = None,
        channel: Optional[Channel] = None,
        action_by_id: Optional[int] = None,
        action_by: Optional[User] = None,
        data: Optional[dict] = None,
        user_id: Optional[int] = None,
        user: Optional[User] = None,
        query: Optional[str] = None,
    ):
        super().__init__(
            app=app,
            type=type or EventType.MESSAGE,
            community_id=community_id,
            community=community,
            group_id=group_id,
            group=group,
            channel_id=channel_id,
            channel=channel,
            action_by_id=action_by_id,
            action_by=action_by,
            data=data,
            user_id=user_id,
            user=user,
        )
        self.query = query

    def from_json(self, data: JSONDict) -> "InlineQueryEvent":
        super().from_json(data)
        if self.data is not None:
            self.query = InlineQuery.build_from_json(
                self.data.get("inlineQuery"), self.app
            )
            self.user = self.action_by
            self.user_id = self.action_by_id

        self.query.user = self.user
        self.query.user_id = self.user_id

        return self



from typing import Optional
from swibots.utils.types import JSONDict
from swibots.api.common.models import User
from swibots.base.switch_object import SwitchObject
import swibots


class RestrictedUser(SwitchObject):
    def __init__(
        self,
        app: "swibots.App" = None,
        restricted: Optional[bool] = None,
        user_id: Optional[int] = None,
        community_id: Optional[str] = None,
        restricted_till: Optional[int] = None,
        user: Optional[User] = None,
    ):
        super().__init__(app)
        self.restricted = restricted
        self.community_id = community_id
        self.user_id = user_id
        self.user = user
        self.restricted_till = restricted_till

    def to_json(self) -> JSONDict:
        return {
            "communityId": self.community_id,
            "restricted": self.restricted,
            "restrictedTillTimestamp": self.restricted_till,
            "userId": self.user_id,
            "userInfo": self.user.to_json() if self.user else None,
        }

    @classmethod
    def from_json(self, data: JSONDict = None) -> "BanInfo":
        if data is not None:
            self.restricted = data.get("restricted")
            self.community_id = data.get("communityId")
            self.restricted_till = data.get("restrictedTillTimestamp")
            self.user_id = data.get("userId")
            self.user = User.build_from_json(data.get("userInfo"), self.app)
        return self



from .group_methods import GroupMethods
from .permission import PermissionMethods
from .roles import RoleMethods
from .rolemember import RoleMemberMethods
from .ban_user import BanUser
from .unban_user import UnbanUser
from .channel_methods import ChannelMethods
from .community_methods import CommunityMethods
from .restrict_user import RestrictUser
from .deduct_xp import DeductXP
from .instant_messaging import InstantMessagingMethods
from .quest_methods import QuestsMethods


class CommunityMethods(
    CommunityMethods,
    GroupMethods,
    RoleMethods,
    PermissionMethods,
    RoleMemberMethods,
    BanUser,
    UnbanUser,
    ChannelMethods,
    RestrictUser,
    DeductXP,
    QuestsMethods,
    InstantMessagingMethods,
):
    pass



from swibots.utils.types import JSONDict
from typing import Dict, Any, Optional, Union
from .types import Component, Icon


class SearchHolder(Component):
    type = "search_holder"

    def __init__(
        self,
        placeholder: str = "Search..",
        callback_data: Optional[str] = None,
        max_size: bool = None,
    ):
        self.placeholder = placeholder
        self.callback_data = callback_data
        self.max_size = max_size

    def to_json(self) -> Dict[str, Any]:
        data = {
            "type": self.type,
            "placeholder": self.placeholder,
            "mainAxisSize": "max" if self.max_size else "min",
        }
        if self.callback_data:
            data["callbackData"] = self.callback_data
        return data


class SearchBar(Component):
    type = "searchbar"

    def __init__(
        self,
        placeholder: str = "Search",
        value: Optional[str] = None,
        label: Optional[str] = None,
        callback_data: Optional[str] = None,
        right_icon: Union[Icon, str] = Icon(
            "https://img.icons8.com/?size=50&id=KPmthqkeTgDN&format=png"
        ),
        left_icon: Union[Icon, str] = Icon(
            "https://img.icons8.com/?size=50&id=357&format=png",
            "https://img.icons8.com/?size=50&id=357&format=png&color=ffffff",
        ),
    ):
        self.placeholder = placeholder
        self.label = label
        self.value = value

        if isinstance(right_icon, str):
            right_icon = Icon(right_icon)
        self.right_icon = right_icon

        if isinstance(left_icon, str):
            left_icon = Icon(left_icon)
        self.left_icon = left_icon
        self.callback_data = callback_data

    def to_json(self) -> Dict[str, Any]:
        data = {"type": self.type, "placeholder": self.placeholder}
        if self.label:
            data["label"] = self.label
        if self.value:
            data["value"] = self.value
        if self.right_icon:
            data["rightIcon"] = self.right_icon.to_json()
        if self.left_icon:
            data["leftIcon"] = self.left_icon.to_json()
        if self.callback_data:
            data["callbackData"] = self.callback_data
        return data



import swibots
from typing import List

from swibots.api.common.models import User
from swibots.api.chat.models import Organization, OrgApp


class OrganizationMethods:
    """Organization methods.

    This class provides methods for interacting with organizations in a chat application.

    Methods:
        get_organizations: Retrieve a list of organizations based on specified parameters.
        get_organization_by_id: Retrieve details of an organization by its ID.
        get_organization_apps: Retrieve the apps associated with a specific organization.
        get_organization_followers: Retrieve the followers of a specific organization.
    """

    async def get_organizations(
        self: "swibots.ApiClient",
        bot_id: int = None,
        community_id: str = None,
        user_id: str = None,
    ) -> List[Organization]:
        """Get a list of organizations.

        Args:
            bot_id (int, optional): The ID of the bot associated with the organizations.
            community_id (str, optional): The ID of the community associated with the organizations.
            user_id (str, optional): The ID of the user associated with the organizations.

        Returns:
            List[Organization]: A list of Organization objects.
        """
        return await self.chat_service.organization.get_organizations(
            bot_id=bot_id, community_id=community_id, user_id=user_id
        )

    async def get_organization_by_id(
        self: "swibots.ApiClient", id: str
    ) -> List[Organization]:
        """Get organization details by ID.

        Args:
            id (str): The ID of the organization.

        Returns:
            Organization: The Organization object.
        """
        return await self.chat_service.organization.get_organization_by_id(id)

    async def get_organization_apps(self: "swibots.ApiClient", id: str) -> List[OrgApp]:
        """Get apps associated with an organization.

        Args:
            id (str): The ID of the organization.

        Returns:
            OrgApp: The OrgApp object.
        """
        return await self.chat_service.organization.get_organization_apps(id)

    async def get_organization_followers(
        self: "swibots.ApiClient", id: str
    ) -> List[User]:
        """Get followers of an organization.

        Args:
            id (str): The ID of the organization.

        Returns:
            List[User]: A list of User objects representing followers.
        """
        return await self.chat_service.organization.get_organization_followers(id)



from typing import TYPE_CHECKING, Type, TypeVar, Optional, Tuple
import swibots
from swibots.api.chat.models import Message, InlineMarkup
from swibots.api.common.models import EmbeddedMedia

if TYPE_CHECKING:
    from swibots.api import ApiClient


class EditMessage:
    async def edit_message(
        self: "ApiClient",
        message_id: int,
        text: str,
        embed_message: EmbeddedMedia = None,
        inline_markup: InlineMarkup = None,
        **kwargs
    ) -> Message:
        """Edit a message

        Parameters:
            message (``~switch.api.chat.models.Message``): The message to edit

        Returns:
            ``~switch.api.chat.models.Message``: The message

        Raises:
            ``~switch.error.SwitchError``: If the message could not be edited

        This method does the same as :meth:`~switch.api.chat.controllers.MessageController.edit_message`.
        """
        return await self.chat_service.messages.edit_message(
            message_id=message_id,
            text=text,
            embed_message=embed_message,
            inline_markup=inline_markup,
            **kwargs
        )

    edit_message_text = edit_message

    async def edit_media(
        self: "swibots.ApiClient",
        message_id: Optional[int] = None,
        media_id: Optional[int] = None,
        message: Optional[str] = None,
        document: Optional[str] = None,
        thumb: Optional[str] = None,
        inline_markup: InlineMarkup = None,
        progress=None,
        progress_args: Optional[Tuple] = None,
        mime_type: Optional[str] = None,
        file_name: Optional[str] = None,
    ) -> Message:
        """Edit media of message

        Args:
            message_id (Optional[int], optional): message id to edit
            media_id (Optional[int], optional): media id to update
            message (Optional[str], optional): message or caption
            document (Optional[str], optional): path to new media
            thumb (Optional[str], optional): thumbnail path
            inline_markup (InlineMarkup, optional):
            progress (_type_, optional): progress callback
            progress_args (Optional[Tuple], optional): progress callback args
            mime_type (Optional[str], optional): mime type of file
            file_name (Optional[str], optional): file name

        Returns:
            Media
        """
        return await self.chat_service.messages.edit_media(
            message_id=message_id,
            media_id=media_id,
            message=message,
            progress=progress,
            progress_args=progress_args,
            inline_markup=inline_markup,
            file_name=file_name,
            mime_type=mime_type,
            thumb=thumb,
            document=document,
        )



import json
from logging import getLogger

import swibots
from swibots.api.community.events import *
from swibots.base import SwitchRestClient, SwitchWSAsyncClient
from swibots.config import get_config
from swibots.errors import SwitchError
from swibots.types import EventType
from swibots.utils.ws.asyncstomp.async_ws_subscription import AsyncWsSubscription
from swibots.utils.ws.common.ws_message import WsMessage
from .controllers import (
    ChannelController,
    GroupController,
    CommunityController,
    RolesController,
    PermissionController,
    RoleMemberController,
    RestrictController,
    BanController,
    InstantMessagingController,
    QuestsController,
)

Logger = getLogger(__name__)


class CommunityClient(SwitchRestClient):
    def __init__(
        self,
        app: "swibots.App" = None,
        base_url: str = None,
        ws_url: str = None,
    ):
        base_url = base_url or get_config()["COMMUNITY_SERVICE"]["BASE_URL"]
        self._ws_url = ws_url or get_config()["COMMUNITY_SERVICE"]["WS_URL"]
        super().__init__(app, base_url)
        self._channels = None
        self._groups = None
        self._communities = None
        self._quests = None
        self._roles = None
        self._rolemember = None
        self._permission = None
        self._ban = None
        self._restrict = None
        self._ws: SwitchWSAsyncClient = None
        self._started = False
        self._instant_messaging = None

    @property
    def quests(self) -> QuestsController:
        if self._quests is None:
            self._quests = QuestsController(self)
        return self._quests

    @property
    def channels(self) -> ChannelController:
        """Get the channels controller"""
        if self._channels is None:
            self._channels = ChannelController(self)
        return self._channels

    @property
    def instant_messaging(self) -> InstantMessagingController:
        if self._instant_messaging is None:
            self._instant_messaging = InstantMessagingController(self)
        return self._instant_messaging

    @property
    def groups(self) -> GroupController:
        """Get the channels controller"""
        if self._groups is None:
            self._groups = GroupController(self)
        return self._groups

    @property
    def communities(self) -> CommunityController:
        """Get the channels controller"""
        if self._communities is None:
            self._communities = CommunityController(self)
        return self._communities

    @property
    def roles(self) -> RolesController:
        """Gets the roles controller"""
        if self._roles is None:
            self._roles = RolesController(self)
        return self._roles

    @property
    def permission(self) -> PermissionController:
        """Gets the permission controller"""
        if self._permission is None:
            self._permission = PermissionController(self)
        return self._permission

    @property
    def restrict(self) -> RestrictController:
        """Get the restrict controller"""
        if self._restrict is None:
            self._restrict = RestrictController(self)
        return self._restrict

    @property
    def rolemember(self) -> RoleMemberController:
        """Gets the rolemember controller"""
        if self._rolemember is None:
            self._rolemember = RoleMemberController(self)
        return self._rolemember

    @property
    def ban(self) -> BanController:
        """Get the ban controller"""
        if self._ban is None:
            self._ban = BanController(self)
        return self._ban

    @property
    def ws(self) -> SwitchWSAsyncClient:
        if self._ws is None:
            self._ws = SwitchWSAsyncClient(self._ws_url, self.token)
        return self._ws

    async def subscribe(self, endpoint: str, callback=None) -> AsyncWsSubscription:
        if self.user is None:
            raise SwitchError("User is not set")
        if callback is None:
            raise SwitchError("Callback is not set")
        return await self.ws.subscribe(endpoint, callback=callback)

    async def subscribe_to_notifications(self, callback=None) -> AsyncWsSubscription:
        subscription = await self.ws.subscribe(
            f"/topic/bot.community.event.{self.user.id}",
            callback=lambda event: callback(self.__parseEvent(event)),
        )
        return subscription

    def __parseEvent(self, raw_message: WsMessage) -> CommunityEvent:
        try:
            return self._parse_event(raw_message)
        except Exception as er:
            Logger.exception(er)

    def _parse_event(self, raw_message: WsMessage) -> CommunityEvent:
        json_data = json.loads(raw_message.body)
        type = json_data.get("type", "COMMUNITY")
        evt = None
        if type == EventType.COMMUNITY_CHANNEL_CREATE.value:
            evt = self.build_object(ChannelCreatedEvent, json_data)
        elif type == EventType.COMMUNITY_CHANNEL_UPDATE.value:
            evt = self.build_object(ChannelUpdatedEvent, json_data)
        elif type == EventType.COMMUNITY_CHANNEL_DELETE.value:
            evt = self.build_object(ChannelDeletedEvent, json_data)
        elif type == EventType.COMMUNITY_UPDATE.value:
            evt = self.build_object(CommunityUpdatedEvent, json_data)
        elif type == EventType.COMMUNITY_GROUP_CREATE.value:
            evt = self.build_object(GroupCreatedEvent, json_data)
        elif type == EventType.COMMUNITY_GROUP_UPDATE.value:
            evt = self.build_object(GroupUpdatedEvent, json_data)
        elif type == EventType.COMMUNITY_GROUP_DELETE.value:
            evt = self.build_object(GroupDeletedEvent, json_data)
        elif type == EventType.COMMUNITY_USER_BAN.value:
            evt = self.build_object(UserBannedEvent, json_data)
        elif type == EventType.COMMUNITY_MEMBER_JOIN.value:
            evt = self.build_object(MemberJoinedEvent, json_data)
        elif type == EventType.COMMUNITY_MEMBER_LEAVE.value:
            evt = self.build_object(MemberLeftEvent, json_data)
        else:
            evt = self.build_object(CommunityEvent, json_data)

        return evt

    async def start(self):
        """Start the community client"""
        if self.user is None:
            raise SwitchError("User is not set")
        await self.ws.connect()
        self._started = True

    async def stop(self):
        if self._started:
            await self.ws.disconnect()
            self._started = False



from typing import List
import swibots
from swibots.api.chat.models import GroupChatHistory


class GetGroupChatHistory:
    async def get_group_chat_history(
        self: "swibots.ApiClient",
        group_id: str,
        community_id: str,
        user_id: int = None,
        page_limit: int = 100,
        page_offset=0,
    ) -> GroupChatHistory:
        """Get group chat history

        Parameters:
            group_id (``int``): The group id
            limit (``int``, *optional*): The maximum number of messages to return. Defaults to 100.
            offset (``int``, *optional*): The offset. Defaults to 0.

        Returns:
            ``List[~switch.api.chat.models.Message]``: The messages

        Raises:
            ``~switch.error.SwitchError``: If the messages could not be retrieved

        This function does the same as :meth:`~switch.api.chat.controllers.MessageController.get_group_chat_history`.
        """
        return await self.chat_service.messages.get_group_chat_history(
            group_id, community_id, user_id, page_limit, page_offset
        )



from swibots.utils.types import JSONDict
from .types import SwitchObject, Component, Image, Badge
from enum import Enum
from typing import Union, List, Literal
from .Progress import ListTileProgress


class ListViewType(Enum):
    DEFAULT = "default"
    SMALL = "small"
    LARGE = "large"
    COMPACT = "compact"


class ListTile(Component):
    type = "list_tile"

    def __init__(
        self,
        title: str,
        description: str = "",
        subtitle: str = "",
        title_extra: str = "",
        description_extra: str = "",
        subtitle_extra: str = "",
        callback_data: str = "",
        thumb: Union[Image, str] = "",
        badges: List[Badge] = None,
        max_size: bool = None,
        progress: ListTileProgress = None,
    ):
        self.title = title
        self.subtitle = subtitle
        self.description = description
        self.title_extra = title_extra
        self.description_extra = description_extra
        self.subtitle_extra = subtitle_extra
        self.callback_data = callback_data
        if isinstance(thumb, str):
            thumb = Image(thumb)
        self.thumb = thumb
        self.badges = badges
        self.max_size = max_size
        self.progress = progress

    def to_json(self):
        data = {
            "type": self.type,
            "title": self.title,
            "subTitle1": self.title_extra,
            "subTitle2": self.description,
            "subtitle3": self.description_extra,
            "subtitle4": self.subtitle,
            "subTitle5": self.subtitle_extra,
            "callbackData": self.callback_data,
            "mainAxisSize": "max" if self.max_size else "min",
        }
        if self.thumb:
            data["image"] = self.thumb.to_json()
        if self.progress:
            data["progress"] = self.progress.to_json()
        if self.badges:
            data["rightComponents"] = [badge.to_json() for badge in self.badges]
        return data


class SmallListTile(ListTile):
    def __init__(
        self, title: str, icon: Union[str, Image] = None, callback_data: str = None
    ):
        super().__init__(title=title, callback_data=callback_data, thumb=icon)


class ListView(Component):
    type = "list_view"

    def __init__(
        self,
        options: List[Union[ListTile, SmallListTile]],
        view_type: ListViewType = ListViewType.DEFAULT,
        title: str = None,
        right_image: str = None,
        image_callback: str = None,
        max_size: bool = None,
    ):
        self.title = title
        self.options = options
        self.view_type = view_type
        self.right_image = right_image
        self.image_callback = image_callback
        self.max_size = max_size

    def to_json_request(self):
        if self.view_type == ListViewType.COMPACT:
            return {
                "type": self.type,
                "title": self.title,
                "listStyle": self.view_type.value,
                "subTitle": self.right_image,
                "subTitleAction": self.image_callback,
                "lists": [tile.to_json() for tile in self.options],
                "mainAxisSize": "max" if self.max_size else "min",
            }

        options = []

        for opt in self.options:
            opt = opt.to_json()
            opt["listStyle"] = self.view_type.value
            opt["mainAxisSize"] = "max" if self.max_size else "min"
            options.append(opt)

        return options



from typing import Optional
from swibots.utils.types import JSONDict
from swibots.base.switch_object import SwitchObject
import swibots


class Role(SwitchObject):
    def __init__(
        self,
        app: "swibots.App" = None,
        id: Optional[int] = 0,
        community_id: Optional[str] = "",
        colour: Optional[str] = "",
        name: Optional[str] = "",
        members_count: Optional[int] = 0,
    ):
        super().__init__(app)

        self.id = id
        self.name = name
        self.community_id = community_id
        self.colour = colour
        self.members_count = members_count

    def to_json(self) -> JSONDict:
        return {
            "roleId": self.id,
            "roleColour": self.colour,
            "roleName": self.name,
            "noOfMembers": self.members_count,
            "communityId": self.community_id,
        }

    def from_json(self, data: JSONDict) -> "Role":
        if data is not None:
            self.id = data.get("id")
            self.name = data.get("roleName")
            self.colour = data.get("roleColour")
            self.community_id = data.get("communityId")
            self.members_count = data.get("noOfMembers")
        return self



from typing import Optional
from swibots.utils.types import JSONDict
from swibots.base.switch_object import SwitchObject
import swibots


class Group(SwitchObject):
    def __init__(
        self,
        app: "swibots.App" = None,
        id: Optional[str] = None,
        name: Optional[str] = None,
        community_id: Optional[str] = None,
        enabled_free: Optional[bool] = None,
        enabled_public: Optional[bool] = None,
        default_group: Optional[bool] = None,
        is_public: Optional[bool] = None,
        created_by: Optional[str] = None,
        icon: Optional[str] = None,
        group_logo_url: Optional[str] = None,
        allowed_content: Optional[str] = None,
        created_at: Optional[str] = None,
        updated_at: Optional[str] = None,
        is_linked_chat: Optional[bool] = None,
    ):
        super().__init__(app)
        self.id = id
        self.name = name
        self.community_id = community_id
        self.enabled_free = enabled_free
        self.enabled_public = enabled_public
        self.default_group = default_group
        self.is_public = is_public
        self.created_by = created_by
        self.icon = icon
        self.group_logo_url = group_logo_url
        self.allowed_content = allowed_content
        self.created_at = created_at
        self.updated_at = updated_at
        self.is_linked_chat = is_linked_chat

    def to_json(self) -> JSONDict:
        return {
            "groupId": self.id,
            "groupName": self.name,
            "communityId": self.community_id,
            "enabledFree": self.enabled_free,
            "enabledPublic": self.enabled_public,
            "defaultGroup": self.default_group,
            "isPublic": self.is_public,
            "createdBy": self.created_by,
            "icon": self.icon,
            "groupLogoUrl": self.group_logo_url,
            "allowedContent": self.allowed_content,
            "createdAt": self.created_at,
            "updatedAt": self.updated_at,
            "linkedChat": self.is_linked_chat,
        }

    def from_json(self, data: JSONDict) -> "Group":
        if data is not None:
            self.id = data.get("groupId")
            self.name = data.get("groupName")
            self.community_id = data.get("communityId")
            self.enabled_free = data.get("enabledFree")
            self.enabled_public = data.get("enabledPublic")
            self.default_group = data.get("defaultGroup")
            self.is_public = data.get("isPublic")
            self.created_by = data.get("createdBy")
            self.icon = data.get("icon")
            self.group_logo_url = data.get("groupLogoUrl")
            self.allowed_content = data.get("allowedContent")
            self.created_at = data.get("createdAt")
            self.updated_at = data.get("updatedAt")
            self.is_linked_chat = data.get("linkedChat")
        return self

    async def m_enabled(self) -> bool:
        """Check whether instant messaging is enabled in current chat"""
        return await self.app.check_messaging_enabled(
            community_id=self.community_id, group_id=self.id
        )



from typing import Optional
from swibots.base import SwitchObject
from swibots.utils.types import JSONDict


class GameInfo(SwitchObject):
    def __init__(
        self,
        id: Optional[str] = None,
        channel_id: Optional[str] = None,
        community_id: Optional[str] = None,
        user_id: Optional[int] = None,
        group_id: Optional[str] = None,
        bot_id: Optional[int] = None,
        score: Optional[int] = None,
        level: Optional[int] = None,
    ):
        self.id = id
        self.channel_id = channel_id
        self.community_id = community_id
        self.user_id = user_id
        self.group_id = group_id
        self.bot_id = bot_id
        self.score = score
        self.level = level

    def from_json(self, data: JSONDict = None) -> "GameInfo":
        if data is not None:
            self.id = data.get("id")
            self.channel_id = data.get("channelId")
            self.community_id = data.get("communityId")
            self.group_id = data.get("groupId")
            self.user_id = data.get("userId")
            self.bot_id = data.get("botId")
            self.score = data.get("score")
            self.level = data.get("level")
        return self

    def to_json(self) -> JSONDict:
        return {
            "id": self.id,
            "channelId": self.channel_id,
            "groupId": self.group_id,
            "communityId": self.community_id,
            "botId": self.bot_id,
            "score": self.score,
            "userId": self.user_id,
            "level": self.level,
        }



from swibots.utils.types import JSONDict
from .types import Component
from enum import Enum
from typing import List
from logging import getLogger
from swibots.base import SwitchObject

LOG = getLogger(__name__)


class TabBarTile(SwitchObject):
    def __init__(
        self,
        title: str = "",
        callback_data: str = None,
        selected: bool = False,
        selection_icon: str = None,
    ):
        self.title = title
        self.callback_data = callback_data
        self.selected = selected
        self.selection_icon = selection_icon

    def to_json(self):
        return {
            "icon": self.title,
            "selected": self.selected,
            "callbackData": self.callback_data,
            "selectionIcon": self.selection_icon,
        }


class TabBarType(Enum):
    SWIPE = "swipe_tab"
    SEGMENTED = "segmented_tab"
    BUTTON = "button_tab"


class TabBar(Component):
    type = "tab"

    def __init__(
        self,
        tabs: List[TabBarTile],
        bar_type: TabBarType = TabBarType.SWIPE,
        theme_color: str = None,
    ):
        self.bar_type = bar_type
        if not tabs:
            raise ValueError("'tabs' are not provided in TabBar.")
        self.tabs = tabs
        self.theme_color = theme_color

    def to_json(self):
        if not any(tab.selected for tab in self.tabs):
            if self.tabs:
                self.tabs[0].selected = True
            else:
                raise ValueError("No tab is marked as selected!")

        return {
            "type": self.type,
            "tabs": [tab.to_json() for tab in self.tabs],
            "tabBarStyle": self.bar_type.value,
            "tabBarColor": self.theme_color,
        }



from typing import Type, TypeVar, Optional, List
import swibots
from swibots.api.bot.models import GameInfo


class GameMethods:
    async def create_leaderboard(
        self: "swibots.ApiClient",
        user_id: str | int,
        score: int,
        level: str | int = None,
        community_id: str = None,
        bot_id: str | int = None,
    ) -> GameInfo:
        return await self.bots_service.games.create_leaderboard(
            user_id, score, level=level, community_id=community_id, bot_id=bot_id
        )

    async def update_leaderboard(
        self: "swibots.ApiClient",
        id: str,
        user_id: str | int,
        score: int,
        level: int | str = None,
        bot_id: str | int = None,
    ) -> GameInfo:
        return await self.bots_service.games.update_leaderboard(
            id, user_id, score, level=level, bot_id=bot_id
        )

    async def get_global_leaderboard(
        self: "swibots.ApiClient", bot_id: Optional[str | int] = None
    ) -> List[GameInfo]:
        return await self.bots_service.games.get_global_leaderboard(bot_id)

    async def get_community_leaderboard(
        self: "swibots.ApiClient", community_id: str, bot_id: Optional[int | str] = None
    ) -> List[GameInfo]:
        return await self.bots_service.games.get_community_leaderboard(
            community_id, bot_id
        )

    async def get_game_score(
        self: "swibots.ApiClient",
        user_id: str | int,
        community_id: Optional[str] = None,
        bot_id: Optional[str | int] = None,
    ) -> GameInfo:
        return await self.bots_service.games.get_game_score(
            user_id, bot_id=bot_id, community_id=community_id
        )



import json
from typing import TYPE_CHECKING, List, Optional
from swibots.api.common.models import User
from swibots.api.callback import AppPage
from swibots.utils.types import JSONDict
from swibots.base import SwitchObject
from .bot_command import BotCommand
from typing import Dict, Any


class BotWelcome(SwitchObject):
    def __init__(
        self,
        text: Optional[str] = None,
        button: Optional[str] = None,
        command: Optional[str] = None,
        thumb: Optional[str] = None,
    ):
        self.text = text
        self.thumb = thumb
        self.button = button
        self.command = command

    def to_json(self) -> JSONDict:
        return {
            "welcomeImage": self.thumb,
            "welcomeText": self.text,
            "buttonName": self.button,
            "buttonCommand": self.command,
        }

    def from_json(self, data: Dict[str, Any] | None) -> Any:
        if data:
            self.thumb = data.get("welcomeImage")
            self.text = data.get("welcomeText")
            self.button = data.get("buttonName")
            self.command = data.get("buttonCommand")
        return self


class BotInfo(User):
    def __init__(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
        username: Optional[str] = None,
        image_url: Optional[str] = None,
        active: Optional[bool] = None,
        deleted: Optional[bool] = None,
        role_info: Optional[str] = None,
        admin: Optional[bool] = None,
        is_bot: Optional[bool] = None,
        commands: Optional[List[BotCommand]] = None,
        description: Optional[str] = None,
        welcome: BotWelcome = None,
        source_code: Optional[str] = None,
        preview: Optional[AppPage] = None,
    ):
        super().__init__(
            id=id,
            name=name,
            username=username,
            image_url=image_url,
            active=active,
            deleted=deleted,
            role_info=role_info,
            admin=admin,
            is_bot=is_bot,
        )
        self.commands: List[BotCommand] = commands or []
        self.description: Optional[str] = description
        self.source_code = source_code
        self.welcome = welcome
        self.preview = preview

    def to_json_request(self) -> JSONDict:
        data = {
            "commands": [command.to_json() for command in self.commands],
            "description": self.description,
            "sourceCode": self.source_code,
        }
        if self.welcome:
            data.update(self.welcome.to_json())
        if self.preview:
            data["preview"] = self.preview.to_json()
        return data

    def from_json(self, data: Optional[JSONDict] = None) -> "BotInfo":
        super().from_json(data)
        if data is not None:
            self.commands = [
                BotCommand().from_json(x) for x in data.get("commands", [])
            ]
            self.description = data.get("description")
            self.welcome = BotWelcome(
                data.get("welcomeText"),
                data.get("buttonName"),
                data.get("buttonCommand"),
                data.get("welcomeImage"),
            )
            self.source_code = data.get("sourceCode")
            preview = data.get("preview")
            if isinstance(preview, str):
                preview = json.loads(preview)
            self.preview = AppPage.build_from_json(preview)

        return self

    def to_json(self) -> JSONDict:
        data = super().to_json()
        if not data.get("commands"):
            data["commands"] = [x.to_json() for x in self.commands]
        if not data.get("description"):
            data["description"] = self.description
        if self.welcome:
            data.update(self.welcome.to_json())
        if self.preview:
            data["preview"] = self.preview.to_json()
        return data



from typing import Tuple, Dict, Union


def _lstrip_str(in_s: str, lstr: str) -> str:
    """
    Args:
        in_s (:obj:`str`): in string
        lstr (:obj:`str`): substr to strip from left side
    Returns:
        :obj:`str`: The stripped string.
    """
    if in_s.startswith(lstr):
        return in_s[len(lstr) :]
    return in_s


class SwitchError(Exception):
    """
    Base class for Switch errors.
    .. seealso:: :wiki:`Exceptions, Warnings and Logging <Exceptions%2C-Warnings-and-Logging>`
    """

    def __init__(self, message: Union[str, Dict]):
        super().__init__()
        message = str(message)
        msg = _lstrip_str(message, "Error: ")
        msg = _lstrip_str(msg, "[Error]: ")
        msg = _lstrip_str(msg, "Bad Request: ")
        if msg != message:
            # api_error - capitalize the msg...
            msg = msg.capitalize()
        self.message = msg

    def __str__(self) -> str:
        return self.message

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}('{self.message}')"

    def __reduce__(self) -> Tuple[type, Tuple[str]]:
        return self.__class__, (self.message,)


class NetworkError(SwitchError):
    """Base class for exceptions due to networking errors.
    Examples:
        :any:`Raw API Bot <examples.rawapibot>`
    """


class TokenInvalidError(SwitchError): ...


class CancelError(SwitchError):
    pass


class BadRequest(SwitchError):
    pass


class FileTooLarge(SwitchError): ...


class ServerError(NetworkError): ...


class UnAuthorized(SwitchError): ...


class ChatEmpty(BadRequest): ...


class MediaEmpty(BadRequest): ...


class InvalidRouteCall(BadRequest): ...


class UnknownBackBlazeError(BadRequest): ...



from swibots.utils.types import JSONDict
from typing import Dict, Any, Optional, Union, List
from .types import Component, Image, Badge, Icon


class AudioPlayer(Component):
    type = "audio_player"

    def __init__(
        self,
        title: str,
        url: str,
        subtitle: str = "",
        thumb: Union[Image, str] = None,
        id: str = None,
        callback_data: str = None,
        previous_callback: Optional[str] = "",
        next_callback: Optional[str] = "",
        max_size: bool = None,
    ):
        self.title = title
        self.url = url
        self.id = id
        self.subtitle = subtitle
        self.thumb = thumb
        self.callback_data = callback_data
        self.previous_callback = previous_callback
        self.next_callback = next_callback
        self.max_size = max_size

    def to_json(self):
        return {
            "type": self.type,
            "title": self.title,
            "subtitle": self.subtitle,
            "image": self.thumb.to_json(),
            "url": self.url,
            "songId": self.id,
            "onSongUpdate": self.callback_data,
            "previousCallback": self.previous_callback,
            "nextCallback": self.next_callback,
            "mainAxisSize": "max" if self.max_size else "min",
        }


class VideoPlayer(Component):
    type = "video_player"

    def __init__(
        self,
        url: str,
        title: str = "",
        subtitle: str = "",
        id: str = "",
        full_screen: bool = False,
        badges: List[Badge] = None,
        callback_data: str = None,
        max_size: bool = None,
    ):
        self.url = url
        self.title = title
        self.id = id
        self.subtitle = subtitle
        self.full_screen = full_screen
        self.badges = badges
        self.callback_data = callback_data
        self.max_size = max_size

    def to_json(self) -> Dict[str, Any]:
        data = {
            "type": self.type,
            "url": self.url,
            "videoId": self.id,
            "title": self.title,
            "subtitle": self.subtitle,
            "fullScreen": self.full_screen,
            "mainAxisSize": "max" if self.max_size else "min",
        }
        if self.callback_data:
            data["onVideoUpdate"] = self.callback_data
        if self.badges:
            data["badges"] = [badge.to_json() for badge in self.badges]
        return data


class Embed(Component):
    type = "embed"

    def __init__(
        self,
        url: str,
        height: Optional[int] = 0,
        width: Optional[int] = 0,
        expand: Optional[bool] = False,
        full_screen: Optional[bool] = True,
        proxy: Optional[dict] = None,
        landscape: bool = False,
        allow_navigation: bool = False,
        navigation_callback: str = None,
        enable_ads: bool = False,
        view_ratio: int = None,
        **kwargs
    ):
        self.url = url
        self.height = height
        self.width = width
        self.full_screen = full_screen
        self.expand = expand
        self.proxy = proxy
        self.landscape = landscape
        self.view_ratio = view_ratio
        self.enable_ads = enable_ads
        self.allow_navigation = allow_navigation
        self.navigation_callback = navigation_callback
        self.__kwargs = kwargs

    def to_json(self):
        data = {
            "type": self.type,
            "url": self.url,
            "height": self.height,
            "width": self.width,
            "fullScreen": self.full_screen,
            "extraOptions": {**self.__kwargs},
        }
        if self.expand:
            data["expansion"] = "flexible_expansion"
        if self.proxy:
            data["proxy"] = self.proxy
        if self.landscape:
            data["extraOptions"]["orientation"] = "landscape"
        if self.allow_navigation:
            data["extraOptions"]["navigation"] = self.allow_navigation
        if self.view_ratio:
            data["extraOptions"]["viewRatio"] = self.view_ratio
        if self.enable_ads:
            data["extraOptions"]["enableAds"] = self.enable_ads
        if self.navigation_callback:
            data["extraOptions"]["navigationCallback"] = self.navigation_callback
        return data


class FileViewer(Component):
    type = "pdf_viewer"

    def __init__(self, url: str):
        self.url = url

    def to_json(self):
        return {"type": self.type, "url": self.url}


class ShortVideo(Component):
    type = "short_video"

    def __init__(
        self,
        url: str = None,
        icons: List[Icon] = None,
        user_image: str = None,
        user_name: str = None,
        title: str = None,
        description: str = None,
    ):
        self.url = url
        self.icons = icons
        self.user_image = user_image
        self.user_name = user_name
        self.title = title
        self.description = description

    def to_json(self):
        data = {
            "type": self.type,
            "videoUrl": self.url,
            "bottom": [icon.to_json() for icon in self.icons],
            "extraOptions": {
                "userAvatar": self.user_image,
                "userName": self.user_name,
                "description": self.description,
                "title": self.title,
            },
        }
        return data


class FeedPanel(Component):
    type = "feed_panel"

    def __init__(self, feeds, next_callback: str):
        self.feeds = feeds
        self.next_callback = next_callback

    def to_json(self) -> Dict[str, Any]:
        data = {
            "type": self.type,
            "feeds": [feed.to_json() for feed in self.feeds],
            "offsetCallbackData": self.next_callback,
        }
        return data



from typing import Callable, Optional
import swibots
from swibots.bots.filters.filter import Filter
from swibots.utils.types import SCT


class OnCommand:
    def on_command(
        self: "swibots.Client", command: SCT[str], filter: Optional[Filter] = None
    ) -> Callable:
        """Decorator for handling new commands."""

        def decorator(func: Callable) -> Callable:
            if isinstance(self, swibots.Client):
                self.add_handler(
                    swibots.bots.handlers.CommandHandler(command, func, filter)
                )

            return func

        return decorator



from typing import List, Union
import swibots
from swibots.api.chat.models import Message


class GetCommunityMediaFilesByStatus:
    async def get_community_media_files_by_status(
        self: "swibots.ApiClient",
        status: Union[str, List[str]],
        community_id: str = None,
        group_id: str = None,
        channel_id: str = None,
        user_id: int = None,
    ) -> List[Message]:
        """Get community media files by status

        Parameters:
            community_id (``int``): The community id
            status (``str`` | `List[str]`): The status

        Returns:
            ``List[~switch.api.chat.models.Message]``: The messages

        Raises:
            ``~switch.error.SwitchError``: If the messages could not be retrieved

        This function does the same as :meth:`~switch.api.chat.controllers.MessageController.get_community_media_files_by_status`.
        """
        return await self.chat_service.messages.get_community_media_files_by_status(
            community_id=community_id,
            user_id=user_id,
            channel_id=channel_id,
            group_id=group_id,
            status=status,
        )



from typing import Optional, Dict
import swibots
from swibots.api.common.events.event import Event
from swibots.api.community.models.channel import Channel
from swibots.api.community.models.community import Community
from swibots.api.community.models.group import Group
from swibots.api.common.models.user import User
from swibots.api.chat.models.message import Message
from swibots.types import EventType
from swibots.utils.types import JSONDict
from .message_event import MessageEvent


class CommandEvent(MessageEvent):
    """Message event"""

    def __init__(
        self,
        app: "swibots.App" = None,
        type: Optional[EventType] = None,
        community_id: Optional[str] = None,
        community: Optional[Community] = None,
        group_id: Optional[str] = None,
        group: Optional[Group] = None,
        channel_id: Optional[str] = None,
        channel: Optional[Channel] = None,
        action_by_id: Optional[int] = None,
        action_by: Optional[User] = None,
        data: Optional[dict] = None,
        user_id: Optional[str] = None,
        user: Optional[User] = None,
        message: Optional[Message] = None,
        message_id: Optional[int] = None,
        command: Optional[str] = None,
        params: Optional[str] = None,
    ):
        super().__init__(
            app=app,
            type=type or EventType.COMMAND,
            community_id=community_id,
            community=community,
            group_id=group_id,
            group=group,
            channel_id=channel_id,
            channel=channel,
            action_by_id=action_by_id,
            action_by=action_by,
            data=data,
            user_id=user_id,
            user=user,
            message_id=message_id,
            message=message,
        )
        self.command = command
        self.params = params

    def from_json(self, data: JSONDict) -> "CommandEvent":
        super().from_json(data)
        if self.data is not None:
            details = self.data.get("commandDetails")
            self.command = self.data.get("command")
            if not self.command and details:
                self.command = details.get("command")
            self.params = self.data.get("commandParams")
        return self



import swibots
from typing import TYPE_CHECKING, Generic, TypeVar
from swibots.api.api_client import ApiClient
from swibots.api.chat.models import Message
from .bot import Bot
from swibots.api.common.events import Event

EventType = TypeVar("EventType", bound="Event")


class BotContext(Generic[EventType], ApiClient):
    def __init__(self, app: "swibots.App", event: EventType):
        self.event = event
        self.app = app
        self._user = app._user

        # copy the api client

        self._chat_client = app.chat_service
        self._auth_client = app.auth_service
        self._community_client = app.community_service
        self._bot_client = app.bots_service

        self.add_handler = self.app.add_handler
        self.remove_handler = self.app.remove_handler
        self.handlers = self.app.handlers
        self.update_bot_commands = self.app.update_bot_commands
        self.set_bot_commands = self.app.set_bot_commands
        self.delete_bot_commands = self.app.delete_bot_commands



from typing import Optional, List
from swibots.utils.types import JSONDict
from swibots.base.switch_object import SwitchObject
from swibots.api.common.models import User
import swibots


class CommunityMember(SwitchObject):
    def __init__(
        self,
        app: "swibots.App" = None,
        admin: Optional[bool] = False,
        community_id: Optional[str] = None,
        enable_notification: Optional[bool] = None,
        id: Optional[int] = None,
        mute_channels: Optional[List[str]] = None,
        mute_groups: Optional[List[str]] = None,
        mute_notification: Optional[bool] = None,
        mute_period: Optional[str] = None,
        role_info: Optional[dict] = None,
        user_id: Optional[str] = None,
        user: Optional[User] = None,
        username: Optional[str] = None,
        request_status: Optional[str] = None,
        xp: Optional[int] = None,
        xp_spend: Optional[int] = None,
    ):
        super().__init__(app)

        self.admin = admin
        self.id = id
        self.enable_notification = enable_notification
        self.mute_notification = mute_notification
        self.community_id = community_id
        self.user_id = user_id
        self.mute_groups = mute_groups
        self.mute_channels = mute_channels
        self.mute_period = mute_period
        self.role_info = role_info
        self.user = user
        self.username = username
        self.xp = xp
        self.xp_spend = xp_spend
        self.request_status = request_status

    def to_json(self) -> JSONDict:
        return {
            "admin": self.admin,
            "communityId": self.community_id,
            "enableNotificationOnMentionAndPin": self.enable_notification,
            "id": self.id,
            "muteChannels": self.mute_channels,
            "muteGroups": self.mute_groups,
            "muteNotification": self.mute_notification,
            "mutePeriod": self.mute_period,
            "roleInfo": self.role_info,
            "userId": self.user_id,
            "userInfo": self.user.to_json() if self.user else None,
            "userName": self.username,
            "requestStatus": self.request_status,
            "xp": self.xp,
            "xpSpend": self.xp_spend,
        }

    def from_json(self, data: JSONDict = None) -> "CommunityMember":
        if data is not None:
            self.admin = data.get("admin")
            self.community_id = data.get("communityId")
            self.enable_notification = data.get("enableNotificationOnMentionAndPin")
            self.id = int(data.get("id") or 0)
            self.mute_channels = data.get("muteChannels")
            self.mute_groups = data.get("muteGroups")
            self.mute_notification = data.get("muteNotification")
            self.mute_period = data.get("mutePeriod")
            self.role_info = data.get("roleInfo")
            self.user_id = int(data.get("userId") or 0)
            self.username = data.get("userName")
            self.user = User.build_from_json(data.get("userInfo"), self.app)
            self.request_status = data.get("requestStatus")
            self.xp = data.get("xp")
            self.xp_spend = data.get("xp_spend")
        return self


class SearchResultUser(SwitchObject):
    def __init__(
        self,
        app: "swibots.App" = None,
        id: str = None,
        member_id: str = None,
        name: str = None,
        username: str = None,
        image_url: str = None,
        active: bool = None,
        deleted: bool = None,
        profile_colour: str = None,
        bot: bool = None,
    ):
        super().__init__(app)
        self.id = id
        self.member_id = member_id
        self.name = name
        self.username = username
        self.image_url = image_url
        self.active = active
        self.deleted = deleted
        self.profile_colour = profile_colour
        self.bot = bot

    def to_json(self) -> JSONDict:
        return {
            "id": self.id,
            "memberId": self.member_id,
            "name": self.name,
            "username": self.username,
            "imageUrl": self.image_url,
            "active": self.active,
            "deleted": self.deleted,
            "profileColour": self.profile_colour,
            "bot": self.bot,
        }

    def from_json(self, data: JSONDict = None) -> "SearchResultUser":
        if data is not None:
            self.id = data.get("id")
            self.member_id = data.get("memberId")
            self.name = data.get("name")
            self.username = data.get("username")
            self.image_url = data.get("imageUrl")
            self.active = data.get("active")
            self.deleted = data.get("deleted")
            self.profile_colour = data.get("profileColour")
            self.bot = data.get("bot")
        return self



from typing import List, Optional
import swibots
from swibots.api.airdrop.models import Referral


class GetReferrals:
    async def get_referral(
        self: "swibots.ApiClient", tournament_id: str | int, user_id: str | int
    ) -> List[Referral]:
        """Get referrals from tournament_id

        Args:
            tournament_id: The Tournament ID to get referrals for.
            user_id: The User ID

        Returns:
            A list of referrals.
        """
        return await self.airdrop_service.tournament.get_referrals(
            tournament_id, user_id
        )



from typing import List
import swibots
from swibots.api.chat.models import Message


class GetFlagMessages:
    async def get_flag_messages(
        self: "swibots.ApiClient", user_id: int = None
    ) -> List[Message]:
        """Get flagged messages

        Parameters:
            user_id (``int``, *optional*): The user id. Defaults to the current user id.

        Returns:
            ``List[~switch.api.chat.models.Message]``: The flagged messages

        Raises:
            ``~switch.error.SwitchError``: If the flagged messages could not be retrieved
        """
        return await self.chat_service.messages.get_flag_messages(user_id)



from typing import TYPE_CHECKING, List, Optional
import swibots

from swibots.base import SwitchObject
from swibots.utils.types import JSONDict

from .inline_keyboard_button import InlineKeyboardButton

if TYPE_CHECKING:
    from .inline_keyboard_button import InlineKeyboardButton

MAX_ROWS = 10
MAX_COLUMNS = 8


class InlineMarkup(SwitchObject):
    def __init__(
        self,
        inline_keyboard: List[List["InlineKeyboardButton"]] = None,
    ):
        super().__init__()
        if isinstance(inline_keyboard, InlineKeyboardButton):
            inline_keyboard = [[inline_keyboard]]
        self._inline_keyboard = inline_keyboard

    @property
    def inline_keyboard(self) -> List[List["InlineKeyboardButton"]]:
        if self._inline_keyboard is None:
            self._inline_keyboard = []
        return self._inline_keyboard

    def add_row(self, buttons: List["InlineKeyboardButton"]):
        if len(buttons) > 0:
            self.inline_keyboard.append(list(buttons))

    def to_json(self) -> JSONDict:
        if self._inline_keyboard is None:
            return None
        return {
            "inlineKeyboard": [
                [x.to_json() for x in row[:MAX_COLUMNS]]
                for row in self._inline_keyboard[:MAX_ROWS]
            ],
        }

    def to_form_data(self):
        form_data = {}
        for i, kb in enumerate(self._inline_keyboard):
            for j, b in enumerate(kb):
                text_key = "inlineKeyboard.inlineKeyboard[{0}][{1}].text".format(i, j)
                form_data[text_key] = b.text
                url_key = "inlineKeyboard.inlineKeyboard[{0}][{1}].url".format(i, j)
                form_data[url_key] = b.url
                game_key = "inlineKeyboard.inlineKeyboard[{0}][{1}].game".format(i, j)
                form_data[game_key] = b.game
                callback_data_key = (
                    "inlineKeyboard.inlineKeyboard[{0}][{1}].callbackData".format(i, j)
                )
                form_data[callback_data_key] = b.callback_data

        return form_data

    def from_json(self, data: JSONDict) -> "InlineMarkup":
        if data is not None and data.get("inlineKeyboard") is not None:
            self._inline_keyboard = [
                [InlineKeyboardButton.build_from_json(x, self.app) for x in row]
                for row in data.get("inlineKeyboard") or []
            ]
        return self



from typing import Optional, List, Union

from enum import Enum
from swibots.utils.types import JSONDict
from .types import Component, Text, Icon


class ButtonVariant(Enum):
    ROUNDED = "rounded"
    OUTLINED = "outlined"
    DEFAULT = "elevated"


class Button(Component):
    type = "action_button"

    def __init__(
        self,
        text: Union[str, Text],
        icon: Union[str, Icon] = "",
        callback_data: Optional[str] = None,
        color: str = None,
        variant: ButtonVariant = ButtonVariant.DEFAULT,
        max_size: bool = None,
        **kwargs
    ):
        if isinstance(text, str):
            text = Text(text)
        self.text = text
        if isinstance(icon, str):
            icon = Icon(icon)
        self.icon = icon
        self.callback_data = callback_data
        self.color = color
        self.variant = variant
        self.clipboard = kwargs.get("clipboard")
        self.action = kwargs.get("action")
        self.url = kwargs.get("url")
        self.file_name = kwargs.get("downloadFileName")
        self.max_size = max_size

    def to_json(self):
        data = {
            "type": self.type,
            "text": self.text.to_json() if self.text else None,
            "callbackData": self.callback_data if self.callback_data else None,
            "color": self.color,
            "variant": self.variant.value,
            "mainAxisSize": "max" if self.max_size else "min",
        }
        if self.icon:
            data["icon"] = self.icon.to_json()
        if self.url:
            data["url"] = self.url
        if self.action:
            data["action"] = self.action
        if self.file_name:
            data["downloadFileName"] = self.file_name
        if self.clipboard:
            data["clipboard"] = True
        return data


class AdButton(Button):
    """
    AdButton: Show Button with different animation for ads.
    """

    def __init__(self, text: Union[str, Text], callback_data: str, **kwargs):
        super().__init__(
            text=text, action="show_ad", callback_data=callback_data, **kwargs
        )


class DownloadButton(Button):
    def __init__(
        self,
        download_url: str,
        file_name: str,
        text: str | Text = "Download",
        icon: str | Icon = "",
        callback_data: str = None,
    ):
        super().__init__(
            text,
            icon,
            url=download_url,
            file_name=file_name,
            action="download",
            callback_data=callback_data,
        )


class ShareButton(Button):
    def __init__(
        self,
        text: str | Text,
        icon: str | Icon = "",
        share_text: str = "",
        color: str = None,
        variant: ButtonVariant = ButtonVariant.DEFAULT,
        **kwargs
    ):
        super().__init__(text, icon, color=color, variant=variant, **kwargs)
        self.share_text = share_text

    def to_json(self):
        data = super().to_json()
        data.update({"share": True, "shareText": self.share_text})
        return data


class ClipboardButton(Button):
    def __init__(
        self, text: str | Text, icon: str | Icon = "", url: str = "", **kwargs
    ):
        super().__init__(text, icon, clipboard=True, url=url, **kwargs)


class ButtonGroup(Component):
    type = "button_group"

    def __init__(self, buttons: List[Button], max_size: bool = None, flexible=False):
        self.buttons = buttons
        self.max_size = max_size
        self.flexible = flexible

    def to_json(self):
        return {
            "type": self.type,
            "buttons": [button.to_json() for button in self.buttons],
            "mainAxisSize": "max" if self.max_size else "min",
            "flexible": self.flexible,
        }


class StickyHeader(Component):
    type = "sticky_header"

    def __init__(
        self,
        text: str,
        color: str = None,
        callback_data: str = None,
        icon: Icon = None,
    ):
        self.text = text
        self.color = color
        self.callback_data = callback_data
        self.icon = icon

    def to_json(self):
        return {
            "type": self.type,
            "text": self.text,
            "color": self.color,
            "callbackData": self.callback_data,
            "icon": self.icon.to_json() if self.icon else None,
        }



from typing import TYPE_CHECKING, Type, TypeVar, Optional
import swibots
from io import BytesIO
from swibots.types import MediaType
from swibots.api.common.models import Media
from swibots.api.chat.models.inline_markup import InlineMarkup

if TYPE_CHECKING:
    from swibots.api import ApiClient


class UploadMedia:
    async def get_media(self: "ApiClient", media_id: int) -> Media:
        """get media by media id

        Args:
            media_id (int): media id

        Returns:
            `Media`
        """
        return await self.chat_service.media.get_media(media_id)

    async def upload_media(
        self: "ApiClient",
        path: str | BytesIO,
        caption: Optional[str] = None,
        description: Optional[str] = None,
        thumb: Optional[str] = None,
        mime_type: Optional[str] = None,
        media_type: Optional[int] = None,
        callback=None,
        callback_args: Optional[tuple] = None,
        part_size: Optional[int] = None,
        task_count: Optional[int] = None,
        premium: Optional[bool] = False,
        **kwargs,
    ) -> Media:
        """upload a file to get `Media` object.

        Arguments:
            path: The path to the file to upload
            caption: str
            description: str
            mime_type: str
            callback: Callable
            callback_args: tuple

        Returns:
             A Media object representing the uploaded file.
        """
        return await self.chat_service.media.upload_media(
            path,
            caption=caption,
            description=description,
            mime_type=mime_type,
            thumb=thumb,
            media_type=media_type,
            callback=callback,
            callback_args=callback_args,
            part_size=part_size,
            task_count=task_count,
            premium=premium,
            **kwargs,
        )

    async def send_document(
        self: "swibots.ApiClient",
        document: str | BytesIO,
        progress=None,
        progress_args=None,
        inline_markup: InlineMarkup = None,
        *args,
        **kwargs,
    ):
        """Alias of send_media to send as document"""
        return await self.send_media(
            document=document,
            media_type=MediaType.DOCUMENT.value,
            progress=progress,
            progress_args=progress_args,
            inline_markup=inline_markup,
            *args,
            **kwargs,
        )

    async def send_media(
        self: "swibots.ApiClient",
        document: str | BytesIO,
        message: Optional[str] = "",
        community_id: Optional[str] = None,
        group_id: Optional[str] = None,
        channel_id: Optional[str] = None,
        user_id: Optional[int] = None,
        user_session_id: Optional[str] = None,
        media: Optional[Media] = None,
        description: Optional[str] = None,
        caption: Optional[str] = None,
        file_name: Optional[str] = None,
        mime_type: Optional[str] = None,
        media_type: Optional[int] = None,
        thumb: Optional[str] = None,
        reply_to_message_id: Optional[int] = None,
        blocking: Optional[bool] = True,
        progress: Optional[callable] = None,
        progress_args: Optional[tuple] = None,
        inline_markup: Optional[InlineMarkup] = None,
        premium: Optional[bool] = False,
        **kwargs,
    ):
        """Send a media to Chat!

        Arguments:
          message (str): the message
          document (str | BytesIO): file path to upload
          caption (str): the media caption
          description (str): the media description
          thumb (str): file path to use as thumb
          community_id (str): The community ID
          group_id (str) The Group ID
          channel_id (str) Channel ID
          user_id (int) User ID
          user_session_id: Session ID, present if bot is added as channel in the community.
          blocking (bool): whether to block task
          progress (Callable[function]): Progress callback function.
          progress_args (tuple)

        Returns:
          `Message` | `Task`
        """
        return await self.chat_service.messages.send_media(
            document=document,
            message=message,
            community_id=community_id,
            group_id=group_id,
            channel_id=channel_id,
            user_id=user_id,
            user_session_id=user_session_id,
            caption=caption,
            media=media,
            description=description,
            file_name=file_name,
            mime_type=mime_type,
            media_type=media_type,
            thumb=thumb,
            blocking=blocking,
            progress=progress,
            progress_args=progress_args,
            reply_to_message_id=reply_to_message_id,
            inline_markup=inline_markup,
            premium=premium,
            **kwargs,
        )

    async def update_media_info(
        self: "swibots.ApiClient",
        media_id: int,
        thumb_url: Optional[str] = None,
        url: Optional[str] = None,
        media_type: Optional[str] = None,
        media: Media = None,
        file_name: Optional[str] = None,
    ):
        """Update Info of uploaded media

        media_id: int: Media Id
        media: Media
        """
        if not media:
            media = Media(
                app=self.chat_service.app,
                id=media_id,
                description=file_name,
                thumbnail_url=thumb_url,
                media_type=media_type,
                url=url,
            )
        return await self.chat_service.media.update_media(
            media_id=media_id, media=media
        )

    async def delete_media(self: "swibots.ApiClient", media_id: int):
        """Delete existing uploaded media!

        Arguments:
          media_id: media id to delete"""
        return await self.chat_service.media.delete_media(media_id)



from typing import TYPE_CHECKING, Type, TypeVar, Optional
import swibots
from swibots.api.chat.models import Message

if TYPE_CHECKING:
    from swibots.api import ApiClient


class PinMessage:
    async def pin_message(
        self: "ApiClient",
        message: Message | int,
        detail: Optional[str] = None,
        message_type: Optional[str] = None,
    ) -> bool:
        """pin a message

        Parameters:
            message (``~switch.api.chat.models.Message``): The message to edit

        Raises:
            ``~switch.error.SwitchError``: If the message could not be pinned

        This method does the same as :meth:`~switch.api.chat.controllers.PostController.pin_message`.
        """
        return await self.chat_service.posts.pin_message(
            message, detail=detail, message_type=message_type
        )



from typing import Optional
from swibots.utils.types import JSONDict
from swibots.base.switch_object import SwitchObject
import swibots


class BanInfo(SwitchObject):
    def __init__(
        self,
        app: "swibots.App" = None,
        banned: Optional[bool] = None,
        message: Optional[str] = None,
    ):
        super().__init__(app)
        self.banned = banned
        self.message = message

    def to_json(self) -> JSONDict:
        return {"isBanned": self.banned, "message": self.message}

    @classmethod
    def from_json(self, data: JSONDict = None) -> "BanInfo":
        if data is not None:
            self.banned = data.get("isBanned")
            self.message = data.get("message")
        return self



from typing import Type, TypeVar
import swibots
from swibots.api.common.models import User
from swibots.api.chat.models import SessionInfo


class GetUser:
    async def get_user(
        self: "swibots.ApiClient", user_id: int = None, username: str = None
    ) -> User:
        """Get User Info from user_id or username

        Args:
          user_id Optional(`int`): The user id to fetch info.
          username [Optional(`str`)]: The username.

        Raises:
          ValueError: if both user_id and username are provided or both are missing!

        """
        return await self.chat_service.messages.get_user(user_id, username=username)

    async def get_session_info(
        self: "swibots.ApiClient", session_id: str
    ) -> SessionInfo:
        """Get session info from session id

        Args:
            session_id (str): session id to get info.

        Returns:
            SessionInfo:
        """
        return await self.chat_service.posts.get_session_info(session_id=session_id)

    async def get_last_seen(self: "swibots.ApiClient", user_id: int) -> int:
        """Get last seen of user

        Args:
            user_id (int): User ID

        Returns:
            int: timestamp when user was last active.
        """
        return await self.chat_service.posts.get_last_seen(user_id)



from typing import Optional

from swibots.base import SwitchObject
from swibots.utils.types import JSONDict


class InlineKeyboardButton(SwitchObject):
    def __init__(
        self,
        text: Optional[str] = None,
        url: Optional[str] = None,
        callback_data: Optional[str] = None,
        game: Optional[bool] = False,
        app: Optional[bool] = False,
    ):
        super().__init__(None)
        self.text = text
        self.url = url
        self.callback_data = callback_data
        self.game = game
        self.is_app = app

    def to_json(self) -> JSONDict:
        return {
            "text": self.text,
            "url": self.url,
            "callbackData": self.callback_data,
            "game": self.game,
            "app": self.is_app,
        }

    def from_json(self, data: JSONDict) -> "InlineKeyboardButton":
        if data is not None:
            self.text = data.get("text")
            self.url = data.get("url")
            self.callback_data = data.get("callbackData")
            self.game = data.get("game")
            self.is_app = data.get("app")
        return self



import asyncio
from typing import List

from swibots.errors import SwitchError

from .async_ws_subscription import AsyncWsSubscription
from concurrent.futures.thread import ThreadPoolExecutor
from swibots.utils.ws.common import WsFrame
from websocket import (
    create_connection,
    WebSocket,
    WebSocketApp,
    WebSocketConnectionClosedException,
    WebSocketException,
    enableTrace,
)
import logging

VERSIONS = "1.1,1.0"

log = logging.getLogger(__name__)


class AsyncWsClient:
    def __init__(
        self,
        url: str,
    ):
        self.url = url
        self.ws = None
        self._loop = asyncio.get_event_loop()
        self.opened = False
        self.connected = False
        self.counter = 0
        self.tasks: List[asyncio.Task] = []
        self._heartbeatTask = None
        self.subscriptions: dict[str, AsyncWsSubscription] = {}
        self._connect_args = {}
        self._connectCallback = None
        self.errorCallback = None
        self._connectIntents = 0
        self._connectInterval = 1
        self._maxConnectIntents = 0
        self.executor = ThreadPoolExecutor(10)
        self._connecting = False
        self._gracefully_disconnect = False

    async def _start_heartbeat(self):
        elapsed = 0
        while self.connected:
            await asyncio.sleep(0.1)
            elapsed += 0.1
            if elapsed >= 5:
                await self._transmit("\n", {})
                elapsed = 0
        log.debug("Heartbeat stopped")

    async def _on_close(self, ws_app, *args):
        self.connected = False
        self._connecting = False
        if self._gracefully_disconnect:
            return
        log.error("Whoops! Lost connection to " + self.url)
        await self._clean_up()
        if (
            self._maxConnectIntents > 0
            and self._connectIntents >= self._maxConnectIntents
        ):
            log.error("Max connection attempts reached. Aborting.")
            raise SwitchError("Max connection attempts reached. Aborting.")
        await asyncio.sleep(self._connectInterval)
        self._connectIntents += 1
        await self.connect(**self._connect_args)

    async def _on_error(self, ws_app, error, *args):
        log.exception(error)
        await self._clean_up()
        await self._on_close(ws_app, *args)

    async def _on_message(self, message, *args):
        frame = WsFrame.unmarshall_single(message)
        if frame.command != "PONG":
            log.debug("\n<<< " + str(message))
        else:
            log.debug("\n<<< " + frame.command)
        _results = []
        if frame.command == "CONNECTED":
            self.connected = True
            self._connecting = False

            if self._connectIntents > 0:
                log.info("Re-Established connection to the server " + self.url)
            else:
                log.info("connected to server " + self.url)

            self._connectIntents = 0
            # resubscribe
            for sub in self.subscriptions.values():
                await self._start_subscription(sub)

            # if self._connectCallback is not None:
            #     _results.append(await self._send_heartbeat(frame))
        elif frame.command == "MESSAGE":
            subscription = frame.headers["subscription"]

            if subscription in self.subscriptions:
                sub = self.subscriptions[subscription]
                messageID = frame.headers["message-id"]

                async def ack(headers):
                    if headers is None:
                        headers = {}
                    return await self.ack(messageID, subscription, headers)

                async def nack(headers):
                    if headers is None:
                        headers = {}
                    return await self.nack(messageID, subscription, headers)

                frame.ack = ack
                frame.nack = nack

                _results.append(
                    # self._loop.create_task(sub.receive(frame))
                    self._loop.run_in_executor(
                        self.executor,
                        lambda: asyncio.new_event_loop().run_until_complete(
                            sub.receive(frame)
                        ),
                    )
                )
            else:
                info = "Unhandled received MESSAGE: " + str(frame)
                log.debug(info)
                _results.append(info)
        elif frame.command == "ERROR":
            if self.errorCallback is not None:
                _results.append(self.errorCallback(frame))
        elif frame.command not in ["PONG", "RECEIPT"]:
            info = "Unhandled received MESSAGE: " + frame.command
            log.debug(info)
            _results.append(info)

        return _results

    async def _transmit(self, command, headers, body=None):
        try:
            out = l = WsFrame.marshall(command, headers, body)
            if command == "\n":
                l = "PING"
                out = command
            log.debug("\n>>> " + l)
            self.ws.send(out)
        except WebSocketConnectionClosedException as er:
            if "socket is already closed.":
                self.connected = False
                self._gracefully_disconnect = True
                await self._on_close(self.ws)
                return
            raise er
        except OSError:
            return
        except Exception as e:
            log.exception(e)
            await self._on_error(self.ws, e)

    async def connect(
        self,
        login=None,
        passcode=None,
        headers=None,
        connectCallback=None,
        errorCallback=None,
        timeout=0,
        **kwargs,
    ):
        headers = self._set_default_headers(headers)
        await self._connect(
            login, passcode, headers, connectCallback, errorCallback, timeout, **kwargs
        )

    async def _connect(
        self,
        login=None,
        passcode=None,
        headers=None,
        connectCallback=None,
        errorCallback=None,
        timeout=30,
        **kwargs,
    ):
        if self.connected:
            log.error("Already connected to " + self.url)
            return

        if self._connecting:
            log.error("Already connecting to " + self.url)
            return

        try:
            self._connecting = True
            self._connect_args = kwargs
            log.debug("Opening web socket...")
            headers = headers if headers is not None else {}

            headers["accept-version"] = VERSIONS
            headers["heart-beat"] = "10000,10000"
            if login is not None:
                headers["login"] = login
            if passcode is not None:
                headers["passcode"] = passcode
            self._connectCallback = connectCallback
            self.errorCallback = errorCallback

            await self.getWebsocket(headers)
            #            self.ws.connect(self.url, header=headers, timeout=120)
            #            await self._transmit("CONNECT", headers)
            self._loop.run_in_executor(
                self.executor,
                lambda: asyncio.new_event_loop().run_until_complete(
                    self.read_messages(headers)
                ),
            )
            #          Thread(target=self.ws.run_forever).start()

            # self.read_messages(headers))
            # headers['host'] = self.url

            # elapsed time
            elapsed = 0
            while not (self.ws):
                await asyncio.sleep(0.5)
                elapsed += 0.5
                if timeout > 0 and elapsed > timeout:
                    raise Exception("Connection timeout")
            if self._connectIntents > 0:
                log.info(f"Retrying connection to {self.url}")
        # await self._start_heartbeat()

        except Exception as e:
            log.exception(e)
            await self._on_error(self.ws, e)

    async def getWebsocket(self, headers):
#        if self.ws:
 #           return self.ws
        self.ws = create_connection(self.url, header=headers, timeout=420)
        await self._transmit("CONNECT", headers)
        return self.ws

    async def read_messages(self, headers):
        try:
            await self.getWebsocket(headers)

            while self.ws.connected:
                message = self.ws.recv()
                await self._on_message(message)
                await self._transmit("\n", headers)
                self.ws.connect(self.url, header=headers)
        except Exception as e:
            log.exception(e)
            await self._on_error(self.ws, e)
            return
        await self._on_close(self.ws)

    async def disconnect(self, disconnectCallback=None, headers=None):
        headers = self._set_default_headers(headers)
        await self._transmit("DISCONNECT", headers)
        self._gracefully_disconnect = True
        if self.ws:
            self.ws.close()
        await self._clean_up()
        if disconnectCallback is not None:
            disconnectCallback()

    async def _clean_up(self):
        try:
            self.connected = False
            self._connecting = False
            self.opened = False
            if self._heartbeatTask:
                self._heartbeatTask.cancel()
                self._heartbeatTask = None

            # if self._heartbeatTask is not None:
            #     await asyncio.wait_for(self._heartbeatTask, 5)
            # [task.cancel() for task in self.tasks]
            # self.ws = None
            self.tasks = []
            for id in list(self.subscriptions.keys()):
                try:
                    await self.subscriptions[id].unsubscribe()
                except KeyError:
                    pass
        except Exception as e:
            log.exception(e)
            log.debug("Error cleaning up: " + str(e))

    async def send(self, destination, headers=None, body=None):
        headers = self._set_default_headers(headers)
        if body is None:
            body = ""
        headers["destination"] = destination
        return await self._transmit("SEND", headers, body)

    async def subscribe(self, destination, callback=None, headers=None):
        headers = self._set_default_headers(headers)
        if "id" not in headers or headers["id"] is None or headers["id"] == "":
            id = "sub-" + str(self.counter)
            self.counter += 1
        else:
            id = headers["id"]

        sub = AsyncWsSubscription(
            client=self,
            destination=destination,
            callback=callback,
            headers=headers,
            id=id,
        )
        await sub.start()
        #        await sub.start()
        #        await self._start_subscription(sub)
        self.subscriptions[id] = sub
        return sub

    def _set_default_headers(self, headers):
        return headers or {}

    async def _start_subscription(self, subscription: AsyncWsSubscription):
        return await subscription.start()

    async def unsubscribe(self, id):
        del self.subscriptions[id]
        return await self._transmit("UNSUBSCRIBE", {"id": id})

    async def ack(self, message_id, subscription, headers):
        headers = self._set_default_headers(headers)
        headers["message-id"] = message_id
        headers["subscription"] = subscription
        return await self._transmit("ACK", headers)

    async def nack(self, message_id, subscription, headers):
        headers = self._set_default_headers(headers)
        headers["message-id"] = message_id
        headers["subscription"] = subscription
        return await self._transmit("NACK", headers)



from .bot_command import BotCommand, BotSubCommand
from .bot_info import BotInfo, BotWelcome
from .game_info import GameInfo

__all__ = ["BotCommand", "BotInfo", "GameInfo", "BotSubCommand", "BotWelcome"]



import json
from typing import Generic, List, Optional, TypeVar
import swibots
from swibots.utils.types import JSONDict


T = TypeVar("T")


class SwitchObject(Generic[T]):
    def __init__(self, app: "swibots.App" = None, **kwargs):
        self._app = app
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def app(self) -> "swibots.App":
        return self._app

    @classmethod
    def build_from_json(
        cls, data: Optional[JSONDict] = None, app: Optional["swibots.App"] = None
    ) -> Optional[T]:
        if data is None:
            return None
        return cls(app).from_json(data)

    @classmethod
    def build_from_json_list(
        cls, data: Optional[JSONDict], app: Optional["swibots.App"] = None
    ) -> List[T]:
        return [cls.build_from_json(item, app) for item in data]

    def to_json_request(self) -> JSONDict:
        return self.to_json()

    def to_json(self) -> JSONDict:
        return self.__dict__

    def from_json(self, data: Optional[JSONDict]) -> T:
        if data:
            for key, value in data.items():
                setattr(self, key, value)
        return self

    def update(self, data: Optional[JSONDict]) -> T:
        new_data = self.to_json()
        new_data.update(data)
        return self.__class__.build_from_json(new_data, self.app)

    def copy(self):
        """Create a copy of object"""
        return self.__class__.build_from_json(self.to_json(), self.app)

    def __repr__(self) -> str:
        filter_dict = {
            x: (
                y.to_json()
                if hasattr(y, "to_json")
                else (y if isinstance(y, (str, dict, int)) else str(y))
            )
            for x, y in self.__dict__.items()
            if y and x != "_app"
        }
        return f"{self.__class__.__name__} {json.dumps(filter_dict, indent=1)}"



from .channel import Channel
from .community import Community, CommunityHeading
from .group import Group
from .role import Role
from .rolepermission import RolePermission
from .rolemember import RoleMember
from .baninfo import BanInfo
from .community_member import CommunityMember, SearchResultUser
from .restricteduser import RestrictedUser
from .quest import Quest, QuestCategory
from .instantmessaging import InstantMessaging

__all__ = [
    "Channel",
    "Community",
    "Group",
    "Role",
    "RolePermission",
    "RoleMember",
    "BanInfo",
    "CommunityHeading",
    "CommunityMember",
    "RestrictedUser",
    "Quest",
    "QuestCategory",
    "InstantMessaging",
    "SearchResultUser",
]



import swibots
from typing import Optional, List, Dict
from swibots.base import SwitchObject
from swibots.utils.types import JSONDict


class BotCommand(SwitchObject):
    def __init__(
        self,
        command: Optional[str] = None,
        description: Optional[str] = None,
        channel: Optional[bool] = False,
        subcommands: Optional[List["BotSubCommand"]] = None,
    ):
        self.active = None
        self.bot_id = None
        self.command = command
        self.description = description
        self.channel = channel
        self.subcommands = subcommands

    def from_json(self, data: JSONDict) -> "BotCommand":
        if data is not None:
            self.bot_id = data.get("botId")
            self.command = data.get("command")
            self.description = data.get("description") or data.get("commandDescription")
            self.channel = data.get("channel")
            self.active = data.get("active")
            self.subcommands = []

            for subcommand, value in data.get("subCommands", {}).items():
                self.subcommands.append(
                    BotSubCommand(
                        command=subcommand, options=value[0] if value else None
                    )
                )
        return self

    def to_json(self) -> JSONDict:
        subcommands = {}
        for command in self.subcommands or []:
            opts = [command.options] if command.options else []
            subcommands[command.command] = opts
        return {
            "botId": self.bot_id,
            "command": self.command,
            "description": self.description,
            "channel": self.channel,
            "subCommands": subcommands,
        }


class BotSubCommand(SwitchObject):
    def __init__(
        self,
        app: "swibots.App" = None,
        command: Optional[str] = None,
        options: Dict = None,
        **kwargs
    ):
        super().__init__(app, **kwargs)
        self.command = command
        self.options = options

    def to_json(self) -> JSONDict:
        return {self.command: [self.options] if self.options else None}



import swibots
from swibots.base import SwitchRestClient
from swibots.config import get_config
from .controllers import TournamentController


class AirdropClient(SwitchRestClient):
    def __init__(
        self,
        app: "swibots.App" = None,
        base_url: str = None,
    ):
        base_url = base_url or get_config()["AIRDROP_SERVICE"]["BASE_URL"]
        super().__init__(app, base_url)

        self._tournament: TournamentController = None

    @property
    def tournament(self) -> TournamentController:
        """Get the controller"""
        if self._tournament is None:
            self._tournament = TournamentController(self)
        return self._tournament

    def prepare_request_headers(self, headers: dict) -> dict:
        headers = super().prepare_request_headers(headers)
        if self.token is not None:
            headers["authtoken"] = f"{self.token}"
        return headers



import swibots
from .auth import AuthClient
from .chat import ChatClient
from .community import CommunityClient
from .bot import BotClient
from .auth.models import AuthUser
from .airdrop import AirdropClient

from .auth.methods import AuthMethods
from .bot.methods import BotMethods
from .chat.methods import ChatMethods
from .community.methods import CommunityMethods
from .airdrop.methods import AirdropMethods


class ApiClient(AuthMethods, BotMethods, ChatMethods, CommunityMethods, AirdropMethods):
    def __init__(self, **kwargs):
        """Initialize the client"""
        self._token: str = None
        self.__dict__.update(kwargs)
        self._chat_client: ChatClient = None
        self._auth_client: AuthClient = None
        self._community_client: CommunityClient = None
        self._bot_client: BotClient = None
        self._user: AuthUser = None
        self._airdrop_client: AirdropClient = None
        self.initialize()

    def initialize(self):
        """Initialize the client"""
        self.chat_service.initialize()
        self.auth_service.initialize()
        self.community_service.initialize()

    @property
    def user(self) -> AuthUser:
        return self._user

    @user.setter
    def user(self, value: AuthUser):
        self._user = value

    @property
    def token(self) -> str:
        """Get the token"""
        return self._token

    @token.setter
    def token(self, value: str):
        """Set the token"""
        self._token = value

    @property
    def chat_service(self) -> ChatClient:
        """Get the chat client"""
        if self._chat_client is None:
            self._chat_client = ChatClient(self)
        return self._chat_client

    @property
    def auth_service(self) -> AuthClient:
        """Get the auth client"""
        if self._auth_client is None:
            self._auth_client = AuthClient(self)
        return self._auth_client

    @property
    def community_service(self) -> CommunityClient:
        """Get the community client"""
        if self._community_client is None:
            self._community_client = CommunityClient(self)
        return self._community_client

    @property
    def airdrop_service(self) -> AirdropClient:
        """Get the airdrop client"""
        if self._airdrop_client is None:
            self._airdrop_client = AirdropClient(self)
        return self._airdrop_client

    @property
    def bots_service(self) -> BotClient:
        """Get the bot client"""
        if self._bot_client is None:
            self._bot_client = BotClient(self)
        return self._bot_client



from typing import Type, TypeVar
import swibots
from swibots.api.bot.models import BotInfo


class UpdateBotInfo:
    async def update_bot_info(self: "swibots.ApiClient", bot_info: BotInfo) -> BotInfo:
        """Update bot info

        Parameters:
            bot_info (``~switch.api.bot.models.BotInfo``): The bot info to update

        Returns:
            :obj:``~switch.api.bot.models.BotInfo``: The bot info

        This functions does the same as :meth:`~switch.api.bot.controllers.BotController.update_bot_info`.
        """
        return await self.bots_service.bots.update_bot_info(bot_info=bot_info)

    async def set_welcome(
        self: "swibots.ApiClient",
        text: str = None,
        thumb: str = None,
        button: str = None,
        command: str = None,
    ):
        """Set bot welcome message

        Args:
            text (str, optional): welcome text
            thumb (str, optional): welcome image
            button (str, optional): button title
            command (str, optional): button command

        Returns:
            `BotInfo`: updated bot info
        """
        return await self.bots_service.bots.set_welcome(
            text=text, thumb=thumb, button=button, command=command
        )



import logging
from typing import TYPE_CHECKING, Optional, List
from ..models import InstantMessaging
from swibots.api.community.models import Community, CommunityMember
from swibots.api.bot.models import BotInfo
from swibots.api.common.models import User
from urllib.parse import urlencode

if TYPE_CHECKING:
    from swibots.api.community import CommunityClient

log = logging.getLogger(__name__)

BASE_PATH = "/v1/community"


class CommunityController:
    def __init__(self, client: "CommunityClient"):
        self.client = client

    async def create_community(self, community: Community):
        response = await self.client.post(BASE_PATH, data=community.to_json_request())
        return self.client.build_object(Community, response.data.get("result"))

    async def get_community(self, community_id: str = "", username: str = ""):
        """Get a community by id or username"""
        if not (community_id or username):
            raise ValueError("community_id or username must be provided.")

        if community_id and username:
            raise ValueError("community_id and username can't be provided together.")
        if username:
            request_url = f"{BASE_PATH}/communityusername?communityUsername={username}"
        else:
            request_url = f"{BASE_PATH}?communityId={community_id}"
        response = await self.client.get(request_url)
        return self.client.build_object(Community, response.data.get("result"))

    async def update_community(self, community: Community):
        response = await self.client.put(BASE_PATH, data=community.to_json_request())
        return self.client.build_object(Community, response.data.get("result"))

    async def deduct_xp(
        self,
        community_id: str,
        user_id: str,
        xp: int = 0,
        description: Optional[str] = None,
    ) -> bool:
        response = await self.client.get(
            f"{BASE_PATH}/deductXP",
            data={
                "communityId": community_id,
                "description": description,
                "userId": user_id,
                "xp": xp,
            },
        )
        return response.data.get("success", False)

    async def get_community_member(
        self, community_id: str, user_id: int
    ) -> CommunityMember:
        response = await self.client.get(
            f"{BASE_PATH}/user?communityId={community_id}&userId={user_id}"
        )
        return self.client.build_object(CommunityMember, response.data.get("result"))

    async def get_community_members(self, community_id: str) -> List[CommunityMember]:
        response = await self.client.get(
            f"{BASE_PATH}/users?communityId={community_id}"
        )
        members = self.client.build_list(
            CommunityMember, response.data.get("result", {}).get("communityMembers")
        )
        userInfo = {
            int(x["id"]): x for x in response.data.get("result", {}).get("userInfoList")
        }
        for community_member in members:
            if data := userInfo.get(community_member.user_id):
                community_member.user = self.client.build_object(User, data)

        return members

    # region

    async def approve_chat_join(
        self,
        members: Optional[int] = None,
        channel_id: Optional[str] = None,
        community_id: Optional[str] = None,
        group_id: Optional[str] = None,
        decline: Optional[bool] = None,
    ):
        await self.client.get(
            f"{BASE_PATH}/members/private/accept",
            data={
                "channelId": channel_id,
                "communityId": community_id,
                "decline": decline,
                "groupId": group_id,
                "members": members,
            },
        )

    # endregion

    # region
    # Commands

    async def get_active_commands(
        self, community_id: str, channel_id: str = None, group_id: Optional[str] = None
    ) -> List[BotInfo]:
        data = {
            "channelId": channel_id or group_id,
            "isGroup": bool(group_id),
            "communityId": community_id,
        }
        response = await self.client.get(
            f"{BASE_PATH}/activecommands?{urlencode(data)}"
        )
        response_data = response.data.get("result", {}).get("channelOrGroup", {})
        return self.client.build_list(BotInfo, response_data.get("bots"))

    # endregion

    # region

    async def is_community_member(self, community_id: str, user_id: int) -> bool:
        response = await self.client.get(
            f"{BASE_PATH}/validate/user/member?communityId={community_id}&user_id={user_id}"
        )
        return response.data.get("result", {}).get("member")

    # endregion

    # region
    async def get_messaging_bots(
        self,
        community_id: str,
        channel_id: Optional[str] = None,
        group_id: Optional[str] = None,
        bot_id: Optional[str] = None,
    ) -> List[InstantMessaging]:
        data = {
            "botId": bot_id,
            "communityId": community_id,
            "channelId": channel_id,
            "groupId": group_id,
        }
        response = await self.client.get(
            f"{BASE_PATH}/instant/messaging/bots?{urlencode(data)}"
        )
        return self.client.build_list(InstantMessaging, response.data.get("result"))

    # endregion

    async def add_premium_member(
        self,
        user_id: str,
        community_id: str,
        channel_id: str = None,
        group_id: str = None,
    ):
        data = {"userId": user_id, "communityId": community_id}
        if channel_id:
            data["channelId"] = channel_id
        if group_id:
            data["groupId"] = group_id
        response = await self.client.post(
            f"{BASE_PATH}/premium/member?{urlencode(data)}"
        )
        return response.data.get("success")

    async def delete_premium_member(
        self,
        user_id: str,
        community_id: str,
        channel_id: str = None,
        group_id: str = None,
    ):
        data = {"userId": user_id, "communityId": community_id}
        if channel_id:
            data["channelId"] = channel_id
        if group_id:
            data["groupId"] = group_id
        response = await self.client.delete(
            f"{BASE_PATH}/premium/member?{urlencode(data)}"
        )
        return response.data.get("success")

    async def get_premium_members(
        self,
        user_id: str,
        community_id: str,
        channel_id: str = None,
        group_id: str = None,
    ):
        data = {"userId": user_id, "communityId": community_id}
        if channel_id:
            data["channelId"] = channel_id
        if group_id:
            data["groupId"] = group_id
        response = await self.client.get(
            f"{BASE_PATH}/premium/member?{urlencode(data)}"
        )
        return response.data



import logging
from typing import TYPE_CHECKING, List

from swibots.api.community.models import Channel
from swibots.errors import SwitchError

if TYPE_CHECKING:
    from swibots.api.community import CommunityClient

log = logging.getLogger(__name__)

BASE_PATH = "/v1/community/channel"


class ChannelController:
    def __init__(self, client: "CommunityClient"):
        self.client = client

    async def get_channel(self, channel_id: str):
        """Get a channel by id"""
        response = await self.client.get(f"{BASE_PATH}?channelId={channel_id}")
        return self.client.build_object(Channel, response.data.get("result"))

    async def create_channel(self, channel: Channel) -> str:
        response = await self.client.post(BASE_PATH, data=channel.to_json_request())
        return response.data.get("result", {}).get("channelId")

    async def delete_channel(self, channel_id: str) -> bool:
        response = await self.client.delete(f"{BASE_PATH}/{channel_id}")
        if response.status_code == 200:
            return True
        logging.debug(response.data)
        if err := response.data.get("errorMessage"):
            raise SwitchError(err)
        return False

    async def update_channel(self, channel: Channel) -> str:
        response = await self.client.put(BASE_PATH, data=channel.to_json_request())
        return response.data.get("result", {}).get("channelId")

    async def get_all_channels(self, community_id: str) -> List[Channel]:
        response = await self.client.get(f"{BASE_PATH}/all?communityId={community_id}")
        return self.client.build_list(Channel, response.data.get("result"))



import swibots

from datetime import datetime
from swibots.api.community.models import Channel


class RestrictUser:
    """
    Class for restricting users in a community.

    Args:
        api_client (swibots.ApiClient): The API client to use.

    Methods:
        restrict_user(community_id, restricted, user_id, until_date):
            Restricts a user in a community.
        update_restricted_user(community_id, restricted, user_id, until_date):
            Updates the restriction status of a user in a community.
    """

    async def restrict_user(
        self: "swibots.ApiClient",
        community_id: str,
        restricted: bool,
        user_id: int,
        until_date: int = 0,
    ) -> bool:
        """
        Restricts a user in a community.

        Args:
            community_id (str): The ID of the community.
            restricted (bool): Whether the user should be restricted.
            user_id (int): The ID of the user to restrict.
            until_date (int): The timestamp until which the user should be restricted.

        Returns:
            bool: Whether the restriction was successful.
        """

        return await self.community_service.restrict.restrict_user(
            community_id=community_id,
            restricted=restricted,
            user_id=user_id,
            until_date=until_date,
        )

    async def update_restricted_user(
        self: "swibots.ApiClient",
        community_id: str,
        restricted: bool,
        user_id: int,
        until_date: int = 0,
    ) -> bool:
        """
        Updates the restriction status of a user in a community.

        Args:
            community_id (str): The ID of the community.
            restricted (bool): Whether the user should be restricted.
            user_id (int): The ID of the user to restrict.
            until_date (float | int): The date until which the user should be restricted.

        Returns:
            bool: Whether the restriction update was successful.
        """

        return await self.community_service.restrict.update_restricted_user(
            community_id, restricted, user_id, until_date
        )

    async def get_restricted_user(
        self: "swibots.ApiClient", community_id: str, user_id: int
    ):
        """get restricted user from community_id and user_id"""
        return await self.community_service.restrict.get_restricted_user(
            community_id, user_id
        )

    async def list_restricted_users(self: "swibots.ApiClient", community_id: str):
        """
           list the restricted users in the community.
           It include previously restricted and unrestricted user too

        Args:
            community_id (str): The ID of the community.

        Returns:
            List[RestrictedUser]: List of restricted users.
        """
        return await self.community_service.restrict.list_restricted_users(community_id)



from swibots.utils.types import JSONDict
from .types import Component, Icon, Text
from typing import Union


class FAB(Component):
    type = "fab"

    def __init__(
        self,
        icon: Union[Icon, str],
        name: Union[Text, str] = None,
        color: str = None,
        callback_data: str = None,
    ):
        if isinstance(name, str):
            name = Text(name)
        self.name = name
        if isinstance(icon, str):
            self.icon = Icon(icon)
        self.color = color
        self.callback_data = callback_data

    def to_json(self):
        data = {
            "type": self.type,
            "icon": self.icon.to_json(),
            "text": self.name.to_json(),
        }
        if self.color:
            data["colour"] = self.color
        if self.callback_data:
            data["callbackData"] = self.callback_data
        return data
